'''
This script will run the chambers identification/segmentation U-Net. 

To see how to extract chamber images with this segmentation masks, see the 
preprocessing section/cell of pipeline.py and getChamberboxes() in utilities.py

@author: jblugagne
'''
from data import saveResult_seg, predictGenerator_seg, postprocess
from model import unet_chambers
from os import listdir

# Files:
DeLTA_data = 'C:/DeepLearning/DeLTA_data/'
inputs_folder = DeLTA_data + 'mother_machine/evaluation/sequence/' # run bioformats2sequence.py first
outputs_folder = DeLTA_data + 'mother_machine/evaluation/chambers_masks/'
model_file = DeLTA_data + 'mother_machine/models/chambers_id_tessiechamp.hdf5'
input_files = listdir(inputs_folder)


# Parameters:
target_size = (512, 512)
input_size = target_size + (1,)

# Load up model:
model = unet_chambers(input_size = input_size)
model.load_weights(model_file)


# Predict:
predGene = predictGenerator_seg(inputs_folder, files_list=input_files , target_size = target_size)
results = model.predict_generator(predGene,len(input_files),verbose=1)

# Post process results:
results[:,:,:,0] = postprocess(results[:,:,:,0])

# Save to disk:
saveResult_seg(outputs_folder,results,files_list=input_files)


# -*- coding: utf-8 -*-
"""
Created on Fri Feb  5 15:42:05 2021

@author: jblugagne
"""

# If running into OOM issues or having trouble with cuDNN loading, try setting
# memory_growth_limit to a value in MB: (eg 1024, 2048...)
memory_growth_limit = None

if memory_growth_limit is not None:
    import tensorflow as tf
    gpus = tf.config.experimental.list_physical_devices('GPU')
    if gpus:
    # Restrict TensorFlow to only allocate 1GB of memory on the first GPU
        try:
            tf.config.experimental.set_virtual_device_configuration(
                gpus[0],
                [tf.config.experimental.VirtualDeviceConfiguration(memory_limit=memory_growth_limit)])
            logical_gpus = tf.config.experimental.list_logical_devices('GPU')
            print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        except RuntimeError as e:
            # Virtual devices must be set before GPUs have been initialized
            print(e)

# Path to model files:
# Download latest models first:
# https://drive.google.com/drive/folders/1nTRVo0rPP9CR9F6WUunVXSXrLNMT_zCP
model_file_chambers = 'C:/DeepLearning/DeLTA_data/mother_machine/models/chambers_id_tessiechamp.hdf5'
model_file_seg = 'C:/DeepLearning/DeLTA_data/mother_machine/models/unet_moma_seg_multisets.hdf5'
model_file_track = 'C:/DeepLearning/DeLTA_data/mother_machine/models/unet_moma_track_multisets.hdf5'

# Model parameters:
target_size_chambers = (512, 512)
target_size_seg = (256, 32)

# Training sets:
training_set_chambers = 'C:/DeepLearning/DeLTA_data/mother_machine/training/chambers_seg_set/train'
training_set_seg = 'C:/DeepLearning/DeLTA_data/mother_machine/training/segmentation_set/train_multisets/'
training_set_track = 'C:/DeepLearning/DeLTA_data/mother_machine/training/tracking_set/train_multisets/'

# Misc:
whole_frame_drift = False # Whether to use the whole frame for drift correction
min_chamber_area = 500 # Minimum size of chambers in pixels for area filtering (0 to +Inf)
min_cell_area = None # Minimum area of cells *IN THE 256x32 SEGMENTATION OUTPUT MASK OF THE UNET*
write_mp4_movies = True # Whether to save MP4 movies showing the DeLTA segmentation & tracking results
rotation_correction = True # Whether to perform automated image rotation correction



'''
This script trains the cell segmentation U-Net

@author: jblugagne
'''
from model import unet_seg
from data import trainGenerator_seg
from tensorflow.keras.callbacks import ModelCheckpoint, EarlyStopping
import os
import config as cfg

# Parameters:
batch_size = 10
epochs = 200
steps_per_epoch = 250
patience = 30

#Data generator parameters:
data_gen_args = dict(rotation = 0.5,
                     shiftX=.05,
                     shiftY=.05,
                     zoom=.15,
                     horizontal_flip=True,
                     histogram_voodoo=True,
                     illumination_voodoo=True)

# Generator init:
myGene = trainGenerator_seg(batch_size,
                            os.path.join(cfg.training_set_seg,'img'),
                            os.path.join(cfg.training_set_seg,'seg'),
                            os.path.join(cfg.training_set_seg,'wei'),
                            augment_params = data_gen_args,
                            target_size = cfg.target_size_seg)

# Define model:
model = unet_seg(input_size = cfg.target_size_seg+(1,))
model.summary()

# Callbacks:
model_checkpoint = ModelCheckpoint(cfg.model_file_seg, 
                                    monitor='loss',
                                    verbose=2,
                                    save_best_only=True)
early_stopping = EarlyStopping(monitor='loss',
                               mode='min',
                               verbose=2,
                               patience=patience)

# Train:
history = model.fit_generator(myGene,
                              steps_per_epoch=steps_per_epoch,
                              epochs=epochs,
                              callbacks=[model_checkpoint, early_stopping])


import unittest
from SyMBac.general_drawing import generate_curve_props, gen_cell_props_for_draw, get_space_size
from SyMBac.phase_contrast_drawing import *
from SyMBac.PSF import get_phase_contrast_kernel, get_condensers
from SyMBac.misc import get_sample_images
import pymunk
from tqdm import tqdm

import importlib

if importlib.util.find_spec("cupy") is None:
    manual_update = True
else:
    manual_update = False


class TestExample(unittest.TestCase):

    def test_partial_example(self):
        real_image = get_sample_images()["E. coli 100x"]

        condensers = get_condensers()
        W, R, diameter = condensers["Ph3"]
        radius = 50
        # F = 5
        λ = 0.75
        resize_amount = 3
        pix_mic_conv = 0.0655  ##0.108379937 micron/pix for 60x, 0.0655 for 100x
        scale = pix_mic_conv / resize_amount
        min_sigma = 0.42 * 0.6 / 6 / scale  # micron#
        sigma = min_sigma
        NA = 1.45
        n = 1.4
        kernel_params = (R, W, radius, scale, NA, n, sigma, λ)

        sim_length = 5
        cell_timeseries, space = run_simulation(
            trench_length=15,
            trench_width=1.5,
            cell_max_length=6,  # 6, long cells # 1.65 short cells
            cell_width=1,  # 1 long cells # 0.95 short cells
            sim_length=sim_length,
            pix_mic_conv=pix_mic_conv,
            gravity=0,
            phys_iters=20,
            max_length_var=3,
            width_var=0.3,
            save_dir="/tmp/"
        )  # growth phase
        main_segments = get_trench_segments(space)
        ID_props = generate_curve_props(cell_timeseries)
        cell_timeseries_properties = Parallel(n_jobs=-1)(
            delayed(gen_cell_props_for_draw)(a, ID_props) for a in tqdm(cell_timeseries, desc='Timeseries Properties'))
        do_transformation = True
        offset = 30
        mask_threshold = 12
        label_masks = True
        space_size = get_space_size(cell_timeseries_properties)
        scenes = Parallel(n_jobs=13)(delayed(draw_scene)(
            cell_properties, do_transformation, space_size, offset, label_masks) for cell_properties in
                                     tqdm(cell_timeseries_properties, desc='Scene Draw:'))

        media_multiplier = 30
        cell_multiplier = 1
        device_multiplier = -50
        y_border_expansion_coefficient = 2
        x_border_expansion_coefficient = 2

        temp_expanded_scene, temp_expanded_scene_no_cells, temp_expanded_mask = generate_PC_OPL(
            main_segments=main_segments,
            offset=offset,
            scene=scenes[0][0],
            mask=scenes[0][1],
            media_multiplier=media_multiplier,
            cell_multiplier=cell_multiplier,
            device_multiplier=cell_multiplier,
            y_border_expansion_coefficient=y_border_expansion_coefficient,
            x_border_expansion_coefficient=x_border_expansion_coefficient,
            fluorescence=False,
            defocus=30
        )

        temp_kernel = get_phase_contrast_kernel(*kernel_params);
        convolved = convolve_rescale(temp_expanded_scene, temp_kernel, 1 / resize_amount, rescale_int=True);
        real_resize, expanded_resized = make_images_same_shape(real_image, convolved, rescale_int=True);

        self.assertIsInstance(space, pymunk.space.Space)
        self.assertEqual(real_resize.shape[0], expanded_resized.shape[0])
        self.assertEqual(real_resize.shape[1], expanded_resized.shape[1])


if __name__ == "__main__":
    unittest.main()



import pyglet
import pymunk
from SyMBac.cell import Cell
from copy import deepcopy
import pickle
def create_space():
    return pymunk.Space()

def update_cell_lengths(cells):
    for cell in cells:
        cell.update_length()


def update_pm_cells(cells):
    for cell in cells:
        if cell.is_dividing():
            daughter_details = cell.create_pm_cell()
            if len(daughter_details) > 2: # Really hacky. Needs fixing because sometimes this returns cell_body, cell shape. So this is a check to ensure that it's returing daughter_x, y and angle
                daughter = Cell(**daughter_details)
                cell.daughter = daughter
                cells.append(daughter)
        else:
            cell.create_pm_cell()

def update_cell_positions(cells):
    for cell in cells:
        cell.update_position()

def wipe_space(space):
    for body, poly in zip(space.bodies, space.shapes):
        if body.body_type == 0:
            space.remove(body)
            space.remove(poly)        



def step_and_update(dt, cells, space, phys_iters, ylim, cell_timeseries,x,sim_length,save_dir):
    for shape in space.shapes:
        if shape.body.position.y < 0 or shape.body.position.y > ylim:
            space.remove(shape.body, shape)
    #new_cells = []
    for cell in cells:
        if cell.shape.body.position.y < 0 or cell.shape.body.position.y > ylim:
            cells.remove(cell)
        else:
            pass
            #new_cells.append(cell)
    #cells = deepcopy(new_cells)

    wipe_space(space)

    update_cell_lengths(cells)
    update_pm_cells(cells)

    for _ in range(phys_iters):
        space.step(dt)
    update_cell_positions(cells)

    #print(str(len(cells))+" cells")
    if x[0] > 1:
        cell_timeseries.append(deepcopy(cells))
    if x[0] == sim_length-1:
        with open(save_dir+"/cell_timeseries.p", "wb") as f:
            pickle.dump(cell_timeseries, f)
        with open(save_dir+"/space_timeseries.p", "wb") as f:
            pickle.dump(space, f)
        pyglet.app.exit()
        return cells
    x[0] += 1
    return (cells)





'''
This file contains functions and class definitions that are used in pipeline.py
A lot of functions are redundant from data.py, but we keep the files separate
to minimize the risk of unforeseen bugs.

@author: jblugagne
'''
import config as cfg
import cv2, os, re
import numpy as np
from scipy.io import savemat
import ffmpeg
import matplotlib.pyplot as plt
import matplotlib.cm as cm


def deskew(image):
    '''
    Compute rotation angle of chambers in image for rotation correction.
    From: https://gist.github.com/panzerama/beebb12a1f9f61e1a7aa8233791bc253
    Not extensively tested. You can skip rotation correction if your chambers
    are about +/- 1 degrees of horizontal

    Parameters
    ----------
    image : 2D numpy array
        Input image.

    Returns
    -------
    rotation_number : float
        Rotation angle of the chambers for correction.

    '''
    
    from skimage.filters import gaussian, threshold_otsu
    from skimage.feature import canny
    from skimage.transform import probabilistic_hough_line


    # canny edges in scikit-image
    edges = canny(image)

    # hough lines
    hough_lines = probabilistic_hough_line(edges)

    # hough lines returns a list of points, in the form ((x1, y1), (x2, y2))
    # representing line segments. the first step is to calculate the slopes of
    # these lines from their paired point values
    slopes = [(y2 - y1)/(x2 - x1) if (x2-x1) else 0 for (x1,y1), (x2, y2) in hough_lines]

    # it just so happens that this slope is also y where y = tan(theta), the angle
    # in a circle by which the line is offset
    rad_angles = [np.arctan(x) for x in slopes]

    # and we change to degrees for the rotation
    deg_angles = [np.degrees(x) for x in rad_angles]

    # which of these degree values is most common?
    histo = np.histogram(deg_angles, bins=180)
    
    # correcting for 'sideways' alignments
    rotation_number = histo[1][np.argmax(histo[0])]

    if rotation_number > 45:
        rotation_number = -(90-rotation_number)
    elif rotation_number < -45:
        rotation_number = 90 - abs(rotation_number)

    return rotation_number

def cropbox(img,box):
    '''
    Crop image

    Parameters
    ----------
    img : 2D numpy array
        Image to crop.
    box : Dictionary
        Dictionary describing the box to cut out, containing the following 
        elements:
            - 'xtl': Top-left corner X coordinate.
            - 'ytl': Top-left corner Y coordinate.
            - 'xbr': Bottom-right corner X coordinate.
            - 'ybr': Bottom-right corner Y coordinate.

    Returns
    -------
    2D numpy array
        Cropped-out region.

    '''
    return img[box['ytl']:box['ybr'],box['xtl']:box['xbr']]

def rangescale(frame, rescale):
    '''
    Rescale image values to be within range

    Parameters
    ----------
    frame : ND numpy array of uint8/uint16/float/bool
        Input image(s).
    rescale : Tuple of 2 values
        Values range for the rescaled image.

    Returns
    -------
    2D numpy array of floats
        Rescaled image

    '''
    frame = frame.astype(np.float32)
    if np.ptp(frame) > 0:
        frame = ((frame-np.min(frame))/np.ptp(frame))*np.ptp(rescale)+rescale[0]
    else:
        frame = np.ones_like(frame)*(rescale[0]+rescale[1])/2
    return frame

def driftcorr(img,template=None,box=None, drift=None):
    '''
    Compute drift between current frame and the reference, and return corrected
    image

    Parameters
    ----------
    img : 2D or 3D numpy array of uint8/uint16/floats
        The frames to correct drfit for.
    template : None or 2D numpy array of uint8/uint16/floats, optional
        The template for drift correction (see getDriftTemplate()).
        default is None.
    box : None or dictionary, optional
        A cropping box to extract the part of the frame to compute drift 
        correction over (see cropbox()).
        default is None.
    drift : None or tuple of 2 numpy arrays, optional
        Pre-computed drift to apply to the img stack. If this is None, you must
        provide a template and box.
        default it None.

    Returns
    -------
    2D/3D numpy array, tuple of len 2
        Drift-corrected image and drift.

    '''
    
    if len(img.shape)==2:
        twoDflag=True
        img = np.expand_dims(img,axis=0)
    else:
        twoDflag=False
        
    if drift is None:
        if template is None: # If we have a position with 0 chambers (see getDriftTemplate)
            return img, (0,0)
        template = rangescale(template,(0,255)).astype(np.uint8) # Making sure its the right format
        xcorr = np.empty([img.shape[0]])
        ycorr = np.empty([img.shape[0]])
    elif twoDflag:
        (xcorr, ycorr) = ([drift[0]],[drift[1]])
    else:
        (xcorr, ycorr) = drift
        
    for i in range(img.shape[0]):
        if drift is None:
            frame = rangescale(img[i],(0,255)).astype(np.uint8) # Making sure its the right format
            driftcorrimg = cropbox(frame,box)
            res = cv2.matchTemplate(driftcorrimg,template,cv2.TM_CCOEFF_NORMED)
            _, _, _, max_loc = cv2.minMaxLoc(res)
            ycorr[i] = max_loc[0] - res.shape[1]/2
            xcorr[i] = max_loc[1] - res.shape[0]/2
        T = np.float32([[1, 0, -ycorr[i]], [0, 1, -xcorr[i]]])
        img[i] = cv2.warpAffine(img[i], T, img.shape[3:0:-1])
    
    if twoDflag:
        return np.squeeze(img), (xcorr[0], ycorr[0])
    else:
        return img, (xcorr, ycorr)


def getChamberBoxes(chambersmask):
    '''
    This function extracts the bounding boxes of the chambers in thebinary mask
    produced by the chambers identification unet

    Parameters
    ----------
    chambersmask : 2D array of uint8/uint16/floats
        The mask of the chambers as returned by the chambers id unet.

    Returns
    -------
    chamberboxes : list of dictionaries
        List of cropping box dictionaries (see cropbox()).

    '''
    chamberboxes = []
    if chambersmask.dtype == bool:
        chambersmask = chambersmask.astype(np.uint8)
    else:
        chambersmask = cv2.threshold(chambersmask,.5,1,cv2.THRESH_BINARY)[1].astype(np.uint8)
    contours = cv2.findContours(chambersmask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[0]
    for chamber in contours:
        xtl, ytl, boxwidth, boxheight = cv2.boundingRect(chamber)
        chamberboxes.append(dict(
                xtl = xtl,
                ytl = int(ytl-(boxheight*.1)), # -10% of height to make sure the top isn't cropped
                xbr = xtl+boxwidth,
                ybr = ytl+boxheight)) # tl = top left, br = bottom right.
    chamberboxes.sort(key=lambda elem: elem['xtl'])# Sorting by top-left X (normally sorted by Y top left)
    return chamberboxes


def getDriftTemplate(chamberboxes,img,whole_frame=False):
    '''
    This function retrieves a region above the chambers to use as drift template

    Parameters
    ----------
    chamberboxes : list of dictionaries
        See getChamberBoxes().
    img : 2D numpy array
        The first frame of a movie to use as reference for drift correction.
    whole_frame : bool, optional
        Whether to use the whole frame as reference instead of the area above
        the chambers.

    Returns
    -------
    2D numpy array or None
        A cropped region of the image to use as template for drift correction.
        If an empty list of chamber boxes is passed, None is returned.
        (see driftcorr()).

    '''
    
    if len(chamberboxes)==0:
        return None
    (y_cut,x_cut) = [round(i*.025) for i in img.shape] # Cutting out 2.5% of the image on eahc side as drift margin
    
    box = dict(
        xtl = x_cut,
        xbr = -x_cut,
        ytl = y_cut,
        ybr = -y_cut if whole_frame else max(chamberboxes,
                                             key=lambda elem: elem['ytl']
                                             )['ytl']-y_cut
        )
        
    return cropbox(img,box)

def getSinglecells(seg):
    '''
    Return masks of single cells

    Parameters
    ----------
    seg : array of uint8/uint16/float/bool
        Mask of cells. Values >0.5 will be considered cell pixels

    Returns
    -------
    singlecellseg : 3D array of uint8
        Stack of single cell masks. Each single-cell mask is stacked along the
        first axis (axis 0)

    '''
    if seg.dtype == bool:
        seg = seg.astype(np.uint8)
    else:
        seg = cv2.threshold(seg,.5,1,cv2.THRESH_BINARY)[1].astype(np.uint8)
    contours = cv2.findContours(seg, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[0]
    contours.sort(key=lambda elem: np.max(elem[:,0,1])) # Sorting along Y
    singlecellseg = np.empty([len(contours),seg.shape[0],seg.shape[1]])
    for c, contour in enumerate(contours):
        singlecellseg[c] = cv2.drawContours(np.zeros((seg.shape[0],seg.shape[1]), dtype=np.uint8),[contour],0, color=1, thickness=-1)
    return singlecellseg

def getTrackingScores(inputs,outputs, method='histogram'):
    '''
    Get overlap scores between input/target cells and tracking outputs

    Parameters
    ----------
    inputs : 2D array of floats
        Segmentation mask of input/target cells that the tracking U-Net is 
        tracking against. (ie segmentation mask of the 'current'/'new' frame)
    outputs : 4D array of floats
        Tracking U-Net output. For each cell in the 'old'/'previous' frame,
        two masks were produced as potential mother and daughter cell outlines.
        (plus a third, unused mask for background)
    method : str, optional
        Method to compute the score, either by binary overlap ('overlap'), by
        element-wise multiplication ('multiplication'), or by weighted 
        histograms ('histogram').
        The default is 'overlap'.

    Returns
    -------
    scores : 3D array of floats
        Overlap scores matrix between tracking predictions and current 
        segmentation mask for each new-old cell, either as a mother or a 
        daughter cell. Cells from the 'old' frame (axis 0) will get a tracking 
        score that corresponds to how much the U-Net tracking output mask 
        overlaps with the segmentation mask of each cell in the 'new' mask 
        (axis 1). Each old cell gets a tracking scores for each new cell as 
        either a potential daughter (mother-daughter relationship, index 1 on 
        axis 2) or as just itself (mother-mother relationship, index 0 on axis
        2)
        Example:
            attrib[1,3,0] == 1 means the 'mother' tracking output of cell #2 in
            the old frame completely overlaps with the mask of cell #4 in the 
            new frame.
            attrib[4,6,1] == 0.7 means the 'daughter' tracking output of cell
            #5 in the old frame overlaps with 70% of the mask of cell #7 in the
            new frame.

    '''

    if method=='overlap':
        targetcells = getSinglecells(inputs).astype(np.uint8)
        scores = np.zeros([outputs.shape[0],targetcells.shape[0],2],dtype=np.float32)
        for o in range(outputs.shape[0]):
            mother_pred = cv2.threshold(outputs[o,:,:,0],.5,1,cv2.THRESH_BINARY)[1].astype(np.uint8)
            daughter_pred = cv2.threshold(outputs[o,:,:,1],.5,1,cv2.THRESH_BINARY)[1].astype(np.uint8)
            for i in range(targetcells.shape[0]):
                scores[o,i,0] = getOverlap(mother_pred,targetcells[i,:,:])
                scores[o,i,1] = getOverlap(daughter_pred,targetcells[i,:,:])

    elif method=='multiplication':
        targetcells = getSinglecells(inputs).astype(np.uint8)
        scores = np.zeros([outputs.shape[0],targetcells.shape[0],2],dtype=np.float32)
        for o in range(outputs.shape[0]):
            for i in range(targetcells.shape[0]):
                scores[o,i,0] = getMult(outputs[o,:,:,0],targetcells[i,:,:])
                scores[o,i,1] = getMult(outputs[o,:,:,1],targetcells[i,:,:])

    elif method=='histogram':
        labels = label_seg(inputs)
        cells = np.max(labels)
        areas, _ = np.histogram(labels, bins=cells+1)
        scores = np.zeros([outputs.shape[0],cells,2],dtype=np.float32)
        for o in range(outputs.shape[0]):
            hist_mother, _ = np.histogram(labels, weights=outputs[o,:,:,0], bins=cells+1)
            hist_daughter, _ = np.histogram(labels, weights=outputs[o,:,:,1], bins=cells+1)
            scores[o,:,0] = hist_mother[1:] / areas[1:]
            scores[o,:,1] = hist_daughter[1:] / areas[1:]

    return scores
        
def getOverlap(output,target):
    '''
    Get portion of tracking output overlapping on target cell

    Parameters
    ----------
    output : array of uint8
        Mask of tracking output.
    target : array of uint8
        Mask of target cell.

    Returns
    -------
    float
        Portion of target cell that is covered by the tracking output.

    '''
    return  cv2.sumElems(
                    cv2.bitwise_and(output,
                                    target)
                )[0]/\
            cv2.sumElems(
                    target
                )[0]

def getMult(output, target):
    '''
    Average tracking output score in target cell mask.

    Parameters
    ----------
    output : array of uint8
        Mask of tracking output.
    target : array of uint8
        Mask of target cell.

    Returns
    -------
    float
        Average tracking output score per target cell pixel.

    '''
    
    return np.sum(np.multiply(output,target))/np.sum(target)

def getAttributions(scores):
    '''
    Get attribution matrix from tracking scores

    Parameters
    ----------
    scores : numpy 3D array of floats
        Tracking scores matrix as produced by the getTrackingScores function.

    Returns
    -------
    attrib : numpy 3D array of bools
        Attribution matrix. Cells from the old frame (axis 0) are attributed to
        cells in the new frame (axis 1). Each new cell can be attributed to an
        old cell as a "mother", ie the same cell (index 0 along axis 2), or as
        a "daughter" of that old cell (index 1 along axis 2)
        Example:
            attrib[1,3,0] == True means cell #4 in the new frame (from top to 
            bottom) is attributed to cell #2 in the old frame.
            attrib[4,6,1] == True means cell #7 in the new frame is 
            attributed as a daughter to cell #5 in the old frame.
        Each 'old' cell can have only 0 or 1 'new' cell identified as itself 
        (ie mother-mother relationship), and 0 or 1 'new' cell identified as a
        daughter (ie mother-daughter relationship). Each 'new' cell can only be
        attributed to 0 or 1 'old' cell, either as a 'mother' or 'daughter'.

    '''
    
    attrib = np.zeros(scores.shape)
    for i in range(scores.shape[1]): # Go over cells in the *new* frame
        if np.sum(scores[:,i,:]>0) == 1: # If only one old cell was attributed to the new cell
            attrib[:,i,:] = scores[:,i,:]>0
        elif np.sum(scores[:,i,:]>0) > 1: # If conflicts
            if np.sum(scores==np.max(scores[:,i,:])) == 1: # One cell has a higher score than the others
                attrib[:,i,:] = scores[:,i,:]==np.max(scores[:,i,:])
            elif np.any(scores[:,i,1]==np.max(scores[:,i,:])): # At least one of the highest scoring old cells would be the mother of this new "daughter"
                attrib[np.argmax(scores[:,i,1]),i,1]=True # keep only the first one (which is also the one higher in the image)
            else: # If only mother-to-mother couplings are in conflict
                attrib[np.argmax(scores[:,i,0]),i,0]=True # keep only the first one (which is also the one higher in the image)
    for o in range(scores.shape[0]): # Go over cells in the *old* frame
        if np.sum(attrib[o,:,0]>0) > 1: # If one old cell gets attributed to more than one new cell (mother-mother coupling)
            tokeep = np.argmax(attrib[o,:,0])
            attrib[o,:,0]=False
            attrib[o,tokeep,0]=True # keep only the first one
        if np.sum(attrib[o,:,1]>0) > 1: # If one old cell gets attributed to more than one new cell (mother-daughter coupling)
            tokeep = np.argmax(attrib[o,:,1])
            attrib[o,:,1]=False
            attrib[o,tokeep,1]=True # keep only the first one
        if np.any(attrib[o,:,0]) and np.any(attrib[o,:,1]) and np.argmax(attrib[o,:,0]) > np.argmax(attrib[o,:,1]): # If somehow the mother is lower than the daughter
            attrib[o,np.argmax(attrib[o,:,0]),1] = True
            attrib[o,np.argmax(attrib[o,:,0]),0] = False
            attrib[o,np.argmax(attrib[o,:,1]),0] = True
            attrib[o,np.argmax(attrib[o,:,1]),1] = False
        if np.any(attrib[o,:,1]) and not np.any(attrib[o,:,0]): # If a cell is attributed a daughter but somehow no mother-mother coupling
            attrib[o,np.argmax(attrib[o,:,1]),0] = True
            attrib[o,np.argmax(attrib[o,:,1]),1] = False
    return attrib

def label_seg(seg,cellnumbers=None):
    '''
    Label cells in segmentation mask

    Parameters
    ----------
    seg : numpy 2D array of float/uint8/uint16/bool
        Cells segmentation mask. Values >0.5 will be considered cell pixels
    cellnumbers : list of ints, optional
        Numbers to attribute to each cell mask, from top to bottom of image.
        Because we are using uint16s, maximum cell number is 65535. If None is 
        provided, the cells will be labeled 1,2,3,... Background is 0
        The default is None.

    Returns
    -------
    label : 2D numpy array of uint16
        Labelled image. Each cell in the image is marked by adjacent pixels 
        with values given by cellnumbers

    '''
    if seg.dtype == bool:
        seg = seg.astype(np.uint8)
    else:
        seg = cv2.threshold(seg,.5,1,cv2.THRESH_BINARY)[1].astype(np.uint8)
    contours = cv2.findContours(seg, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[0]
    contours.sort(key=lambda elem: np.max(elem[:,0,1])) # Sorting along Y
    label = np.zeros(seg.shape,dtype=np.uint16) # Because we use uint16, we can not have >65535 cells in one movie
    for c, contour in enumerate(contours):
        label = cv2.fillPoly(label,[contour],c+1 if cellnumbers is None else cellnumbers[c])
    return label

def updatelineage(newframe, stack, framenb=0, lineage=None, attrib=None):
    '''
    Update lineage list and labels stack based on attribution matrix

    Parameters
    ----------
    newframe : numpy 2D array
        Segmentation output for current frame.
    stack : numpy 3D tensor of uint16
        Labelled stack. Each frame until the current frame contains a numbered 
        mask, with the number corresponding the cell number in the lineage. See
        label_seg()
    framenb : int, optional
        Current frame number (0-based indexing).
        The default is 0.
    lineage : List of dicts, optional
        List of all the cells encountered in the chamber. Each cell dict 
        contains the following keys:
            - mother, int: The number of the mother cell in the lineage 
            (1-based indexing)
            - framenbs, list of ints: The list of frames where the cell is 
            present (1-based indexing)
            - daughters, list of ints: A list of the same length as framenbs,
            with each element equal to the number of the daughter in the 
            lineage (1-based indexing) if a division occured at the 
            corresponding frame number, and equal to 0 otherwise.
        Note: We used 1-based indexing for compatibility with earlier code that
        was written in Matlab.
        If None is provided, the lineage will be initialized and all cells in 
        the newframe image will be given 0 as a mother.
        The default is None.
    attrib : numpy 2D array of bool, optional
        Attribution matrix for the current frame/timepoint as produced by 
        number 0. If None, lineage and stack will not be updated.
        getAttributions(). New cells with no attributions are given mother cell
        The default is None.

    Returns
    -------
    lineage : List of dicts
        lineage updated for the current frame and attrib matrix. See lineage 
        input description above.
    stack : numpy 3D tensor of uint16
        labelled stack updated for the current frame and attrib matrix. See 
        lineage input description above..

    '''
    
    if lineage==None: # Not initialized yet
        lineage = []
        stack[framenb] = label_seg(newframe)
        cells = np.unique(stack[framenb])
        for c in cells[1:]: # Avoiding cell==0 (background)
            lineage.append(createcell(framenb+1)) # We're using framenb +1 to be compatible with old Matlab code
    else: # Update according to attribution matrix
        # Init cell numbers, retrieve previous stack frame:
        cellnumbers = []
        label_old = label_seg((stack[framenb-1]>0).astype(np.uint8))
        
        # If previous stack frame has no cells:
        if not np.any(label_old>0): # This can happen if the chamber is empty and suddenly cells appear (can be a segmentation error, or a focus problem)
            contours = cv2.findContours(newframe.astype(np.uint8), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[0] # In this case just label
            for _ in range(len(contours)):
                lineage.append(createcell(framenb+1))
                cellnumbers.append(len(lineage))
                
        # If previous stack frame has cells:
        else:
            for i in range(attrib.shape[1]): # Go through new cells
                if np.any(attrib[:,i,0]): # If cell is tracked from an old cell: (mother-mother)
                    cellnum = int(stack[framenb-1][label_old==(np.nonzero(attrib[:,i,0])[0]+1)[0]][0]) # retrieving "mother" cell number from old stack frame
                    lineage[cellnum-1]['framenbs'].append(float(framenb+1)) # Update frame number list
                    if len(lineage[cellnum-1]['daughters']) < len(lineage[cellnum-1]['framenbs']): # If framenbs hasn't been updated (in case somehow daughter goes through before mother)
                        lineage[cellnum-1]['daughters'].append(0.)
                    cellnumbers.append(cellnum)
                elif np.any(attrib[:,i,1]): # Cell is tracked as a daughter cell (mother-daughter)
                    mothernb = int(stack[framenb-1][label_old==(np.nonzero(attrib[:,i,1])[0]+1)[0]][0]) # retrieving cell number from stack
                    # Create new cell, as daughter of mother:
                    lineage.append(createcell(framenb+1,mothernb))
                    cellnumbers.append(len(lineage))
                    # Update 'daughters' key of mother cell:
                    if len(lineage[mothernb-1]['daughters']) < len(lineage[mothernb-1]['framenbs']): # This situation should never happen actually
                        lineage[mothernb-1]['daughters'].append(float(cellnumbers[-1]))
                    else:
                        lineage[mothernb-1]['daughters'][-1] = float(cellnumbers[-1])
                else: # Orphan cell
                    lineage.append(createcell(framenb+1))
                    cellnumbers.append(len(lineage))
                    
        # Compile new labeled frame, push to stack:
        stack[framenb] = label_seg(newframe,cellnumbers=cellnumbers)
    return lineage, stack
        

def createcell(framenb, mothernb=0):
    '''
    Create cell to append to lineage list

    Parameters
    ----------
    framenb : int
        Frame that the cell first appears (1-based indexing).
    mothernb : int, optional
        Number of the mother cell in the lineage (1-based indexing).
        The default is 0. (ie unknown mother)

    Returns
    -------
    dict
        Initialized cell dictionary.

    '''
    return dict(framenbs=[float(framenb)], # For compatibility with old version of the pipeline
                mothernb=mothernb,
                daughters=[0.])

def loadmodels(toload=['chambers','segmentation','tracking']):
    '''
    Load models (as specified in config.py)

    Parameters
    ----------
    toload : list of str, optional
        Which of the 3 models to load. 
        The default is ['chambers','segmentation','tracking'].

    Returns
    -------
    models : dict
        Dictionary containing the models specified.

    '''
    
    
    from model import unet_chambers, unet_seg, unet_track
    models = dict()
    
    if 'chambers' in toload:
        models['chambers'] = unet_chambers(input_size = cfg.target_size_chambers + (1,))
        models['chambers'].load_weights(cfg.model_file_chambers)
    
    if 'segmentation' in toload:
        models['segmentation'] = unet_seg(input_size = cfg.target_size_seg + (1,))
        models['segmentation'].load_weights(cfg.model_file_seg)
    
    if 'tracking' in toload:
        models['tracking'] = unet_track(input_size = cfg.target_size_seg + (4,))
        models['tracking'].load_weights(cfg.model_file_track)
    
    return models

def legacysave(res_file, 
               xpfile, 
               delta_res, 
               chamberboxes, 
               rot_corr, 
               driftvalues,
               moviedimensions):
    '''
    Save single position pipeline results in the legacy Matlab format

    Parameters
    ----------
    res_file : str
        Path to save the data to.
    xpfile : str
        Path to the experimental folder.
    res : list
        List of dicts for each chamber produced by delta pipeline.
    chamberboxes : list
        List of dicts describing single chambers positions in the image.
    rot_corr : float
        Rotation applied to images to get the chambers horizontal.
    driftvalues : list
        XY drift values computed by pipeline preprocessing step.

    Returns
    -------
    None.

    '''
    
    if delta_res is not None:
        savemat(res_file,
                {'res': delta_res,
                 'tiffile': xpfile,
                 'moviedimensions': moviedimensions,
                 'proc': {'rotation': rot_corr,
                          'chambers': np.array([[c['xtl'],c['ytl'],c['xbr']-c['xtl'],c['ybr']-c['ytl']] for c in chamberboxes],dtype=np.float64),
                          'XYdrift': np.flip(np.transpose(-np.array(driftvalues)),axis=1)}
                 })
    
    else:
        savemat(res_file,
                {'res': [],
                 'tiffile': xpfile,
                 'moviedimensions': moviedimensions,
                 'proc': {'rotation': rot_corr,
                          'chambers': [],
                          'XYdrift': []}
                 })

def getxppathdialog(ask_folder=False):
    '''
    Pop up window to select experiment file or folder.

    Parameters
    ----------
    ask_folder : bool, optional
        Folder selection window will pop up instead of file selection.
        The default is False.

    Returns
    -------
    file_path : str
        Path to experiment file or folder.

    '''
    
    import tkinter as tk
    from tkinter import filedialog
    
    root = tk.Tk()
    # root.withdraw() # For some reason this doesn't work with askdirectory?
    if ask_folder:
        file_path = filedialog.askdirectory(title='Please select experiment folder(s)',
                                              mustexist=True)
        
    else:
        file_path = filedialog.askopenfilenames(title='Please select experiment files(s)')
    
    root.destroy()
        
    return file_path

def results_movie(xp, 
                  pos, 
                  res, 
                  chamberboxes,
                  rot_corr=0, 
                  driftvalues=None,
                  display=False
                  ):
    '''
    Produce movie illustrating DeLTA results.

    Parameters
    ----------
    xp : xpreader object
        See xpreader() in utilities.py. 
    pos : int
        Position to process.
    res : list
        List of dicts for each chamber produced by delta pipeline.
    chamberboxes : list or None
        List of dicts describing single chambers positions in the image.
    rot_corr : int or float, optional
        Rotation applied to images to get the chambers horizontal.
        The default is 0.
    driftvalues : list or None, optional
        XY drift values computed by pipeline preprocessing step.
        The default is None.
    display : bool, optional
        Flag to display movie frames as they are create.
        The default is False.

    Returns
    -------
    movie : 4D numpy array of floats.
        Array of RGB frames showing DeLTA result. Dimensions are timepoints 
        -by- X -by- Y -by- 3.

    '''
    
    # Get trans frames:
    transframes = xp.getframes(positions=pos,channels=0,rescale=(0,1),rotate=rot_corr) # Load all trans frames, with rotation correction
    transframes, _ = driftcorr(np.squeeze(transframes), drift=driftvalues) # Apply drift correction
    movie = np.empty(transframes.shape+(3,))
    
    for f, frame in enumerate(transframes):
        
        #RGB-ify:
        frame = np.repeat(frame[:,:,np.newaxis],3,axis=-1)
        
        # Add frame number text:
        frame = cv2.putText(frame,
                            text='frame %06d' %(f+1,),
                            org=(int(frame.shape[0]*.05),
                                 int(frame.shape[0]*.97)),
                            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                            fontScale=1,
                            color=(1, 1, 1, 1),
                            thickness=2
                            )
        
        # Run through chambers, add contours & division arrows:
        if chamberboxes is not None:
            for chamber, chamberbox in enumerate(chamberboxes):
                
                # Get chamber-specific variables:
                lin = res[chamber]['lineage']
                labels = res[chamber]['labelsstack_resized'][f]
                colors = getrandomcolors(len(lin), seed=chamber)
                cells, contours = getcellsinframe(labels, get_contours=True)
                
                # Run through cells in labelled frame:
                for c, cell in enumerate(cells):
                    
                    # Draw contours:
                    frame = cv2.drawContours(frame,
                                             contours,
                                             c,
                                             color=colors[cell],
                                             thickness=1,
                                             offset=(chamberbox['xtl'],
                                                     chamberbox['ytl'])
                                             )
                    
                    # Plot division arrows:
                    daughter = int(lin[cell]['daughters'][lin[cell]['framenbs'].index(f+1)]-1)
                    if daughter>0:
                        # Get mother centroid:
                        centroid_mother = getcentroid(contours[c])
                        centroid_mother = (centroid_mother[0]+chamberbox['xtl'],
                                           centroid_mother[1]+chamberbox['ytl'])
                        # Get daughter centroid:
                        centroid_daughter = getcentroid(contours[cells.index(daughter)])
                        centroid_daughter = (centroid_daughter[0]+chamberbox['xtl'],
                                             centroid_daughter[1]+chamberbox['ytl'])
                        # Plot arrow:
                        frame = cv2.arrowedLine(frame,
                                                centroid_mother,
                                                centroid_daughter,
                                                color=(1,1,1),
                                                thickness=1)

        # Show current frame:
        if display:
            plt.imshow(frame)
            plt.show()
        
        # Add to movie array:
        movie[f] = frame
    
    return movie

def getcellsinframe(labels,get_contours=False):
    '''
    Get numbers of cells present in frame, sorted along Y axis

    Parameters
    ----------
    labels : 2D numpy array of ints
        Single frame from labels stack.
    get_contours : bool, optional
        Flag to get cv2 contours.

    Returns
    -------
    cells : list
        Cell numbers (0-based indexing).
    contours : list
        List of cv2 contours for each cell. Returned if get_contours==True.

    '''
    
    cells, ind = np.unique(labels,return_index=True)
    cells = [cell-1 for _,cell in sorted(zip(ind,cells))] # Sorting along Y axis, 1-based to 0-based, & removing 1st value (background)
    cells = [cell for cell in cells if cell!=-1] # Remove background
    
    # Get opencv contours:
    if get_contours:
        contours = []
        for c, cell in enumerate(cells):
            # Have to do it this way to avoid cases where border is shrunk out
            cnt, _ = cv2.findContours((labels==cell+1).astype(np.uint8),
                                       cv2.RETR_EXTERNAL,
                                       cv2.CHAIN_APPROX_SIMPLE)
            contours += cnt
        return cells, contours
    else:
        return cells

def getrandomcolors(num_colors, seed=0):
    '''
    Pseudo-randomly generate list of random hsv colors.

    Parameters
    ----------
    num_colors : int
        Number of desired colors.
    seed : None or int, optional
        Seed to use with numpy.random.seed(). 
        The default is 0.

    Returns
    -------
    colors : list
        List of RGB values (0-1 interval).

    '''
    
    # Get colors:
    cmap = cm.get_cmap('hsv', lut=num_colors)
    colors = [cmap(i)[0:3] for i in range(num_colors)]
    
    # Pseudo randomly shuffle colors:
    if seed is not None:
        np.random.seed(seed)
    np.random.shuffle(colors)
    
    return colors

def getcentroid(contour):
    '''
    Get centroid of cv2 contour

    Parameters
    ----------
    contour : 3D numpy array
        Blob contour generated by cv2.findContours().

    Returns
    -------
    cx : int
        X-axis coordinate of centroid.
    cy : int
        Y-axis coordinate of centroid.

    '''
    
    if contour.shape[0]>2: # Looks like cv2.moments treats it as an image
        # Calculate moments for each contour
        M = cv2.moments(contour)
        # Calculate x,y coordinate of center
        if M["m00"] != 0:
            cx = int(M["m10"] / M["m00"])
            cy = int(M["m01"] / M["m00"])
        else:
            cx, cy = 0, 0
    else:
        cx = int(np.mean(contour[:,:,0]))
        cy = int(np.mean(contour[:,:,1]))
        
    return cx,cy



def vidwrite(images, filename, crf=20, verbose=1):
    '''
    Write images stack to video file with h264 compression.

    Parameters
    ----------
    images : 4D numpy array
        Stack of RGB images to write to video file.
    filename : str
        File name to write video to. (Overwritten if exists)
    crf : int, optional
        Compression rate. 'Sane' values are 17-28. See 
        https://trac.ffmpeg.org/wiki/Encode/H.264
        The default is 23.
    verbose : int, optional
        Verbosity of console output. 
        The default is 1.

    Returns
    -------
    None.

    '''
    
    # Initialize ffmpeg parameters:
    height,width,_ = images[0].shape
    quiet = [] if verbose else ['-loglevel', 'error', '-hide_banner']
    process = (
        ffmpeg
            .input('pipe:', format='rawvideo', pix_fmt='rgb24', s='{}x{}'.format(width, height),r=7)
            .output(filename, pix_fmt='yuv420p', vcodec='libx264', crf=crf, preset='veryslow')
            .global_args(*quiet)
            .overwrite_output()
            .run_async(pipe_stdin=True)
    )
    
    # Write frames:
    for frame in images:
        process.stdin.write(
            frame
                .astype(np.uint8)
                .tobytes()
        )
    
    # Close file stream:
    process.stdin.close()
    
    # Wait for processing + close to complete:
    process.wait()

class xpreader:
    
    def __init__(self, filename, channelnames=None, use_bioformats=False, prototype=None, fileorder='pct', filenamesindexing=1):
        '''
        Initialize experiment reader

        Parameters
        ----------
        filename : String
            Path to experiment tif file or directory.If the path leads to a
            directory, the experiment folder is expected to contain exclusively
            single-page tif images. If no prototype is provided, the filenames 
            are expected to be of the following C-style format: 
            %s%d%s%d%s%d.tif, with the 3 %d digit strings being zero-padded 
            decimal representation of the position, channel and frame/timepoint
            number of each image file.
            Valid examples:
                Pos01_Cha3_Fra0005.tif
                p3c1t034.tif
                xy 145 - fluo 029 - timepoint 005935 .TIFF
        channelnames : List/tuple of strings or None, optional
            Names of the acquisition channels ('trans', 'gfp', ...).
            The default is None.
        use_bioformats : bool, optional
            Flag to use the bioformats reader.
            The default is False.
        prototype: string, optional
            Filename prototype to use when reading single-page tif images from
            a sequence folder, in C-style formatting. Folder separators can be
            used. If None, the prototype will be estimated from the first tif
            file in the folder. For example, an experiment from micromanager 
            can be processed with prototype =
            'Pos%01d/img_channel%03d_position%03d_time%09d_z000.tif'
            and fileorder = 'pcpt' and filenamesindexing = 0
            The default is None.
        fileorder: string, optional
            Order of the numbers in the prototype, with 'p' for positions/
            series, 'c' for imaging channels, and 't' for timepoints/frames.
            For example 'pct' indicates that the first number is going to be
            positions, then channels, then timepoints. You can use the same 
            letter multiple times, like 'pcpt'.
            The default is 'pct'
        filenamesindexing = int
            Selects between 0-based or 1-based indexing in the filename. If 
            1, position 0 will be referenced as position 1 in the filename.
            The default is 1
            

        Raises
        ------
        ValueError
            If the filenames in the experimental directory do not follow the 
            correct format, a ValueError will be raised.

        Returns
        -------
        None.

        '''
        
        _, file_extension = os.path.splitext(filename)
        self.use_bioformats = use_bioformats
        self.channelnames = channelnames
        if use_bioformats:
            import bioformats, javabridge
            javabridge.start_vm(class_path=bioformats.JARS)
            self.filetype = file_extension.lower()[1:]
            self.filehandle = bioformats.ImageReader(filename)
            md = bioformats.OMEXML(bioformats.get_omexml_metadata(path=filename))
            self.positions = md.get_image_count()
            self.timepoints = md.image(0).Pixels.SizeT # Here I'm going to assume all series have the same number of timepoints
            self.channels = md.image(0).Pixels.channel_count
            self.x = md.image(0).Pixels.SizeX
            self.y = md.image(0).Pixels.SizeY
            # Get first image to get datatype (there's probably a better way to do this...)
            self.dtype = self.filehandle.read(rescale=False,c=0).dtype
        
        elif os.path.isdir(filename): # Experiment is stored as individual image TIFF files in a folder
            self.filetype = 'dir'
            self.filehandle = filename
            self.fileorder = fileorder
            self.filenamesindexing = filenamesindexing
            # If filename prototype is not provided, guess it from the first file:
            if prototype is None:
                imgfiles = [x for x in os.listdir(filename) if os.path.splitext(x)[1].lower() in ('.tif','.tiff')]
                # Here we assume all images in the folder follow the same naming convention:
                numstrs = re.findall("\d+", imgfiles[0]) # Get digits sequences in first filename
                charstrs = re.findall("\D+", imgfiles[0]) # Get character sequences in first filename
                if len(numstrs)!=3 or len(charstrs)!=4:
                    raise ValueError('Filename formatting error. See documentation for image sequence formatting')
                # Create the string prototype to be used to generate filenames on the fly:
                # Order is position, channel, frame/timepoint
                self.prototype =    charstrs[0]+'%0'+str(len(numstrs[0]))+'d'+\
                                    charstrs[1]+'%0'+str(len(numstrs[1]))+'d'+\
                                    charstrs[2]+'%0'+str(len(numstrs[2]))+'d'+\
                                    charstrs[3]
            else:
                self.prototype = prototype
            # Get experiment settings by testing if relevant files exist: 
            # Get number of positions:
            if 'p' in self.fileorder:
                self.positions = 0
                while(os.path.exists(self.getfilenamefromprototype(self.positions,0,0))): self.positions+=1
            else:
                self.positions = 1
            # Get number of channels:
            if 'c' in self.fileorder:
                self.channels = 0
                while(os.path.exists(self.getfilenamefromprototype(0,self.channels,0))): self.channels+=1
            else:
                self.channels = 1
            # Get number of frames/timepoints:
            if 't' in self.fileorder:
                self.timepoints = 0
                while(os.path.exists(self.getfilenamefromprototype(0,0,self.timepoints))): self.timepoints+=1
            else:
                self.timepoints = 1 # I guess this shouldn't really happen
            # Get image specs:
            I = cv2.imread(self.getfilenamefromprototype(0,0,0),cv2.IMREAD_ANYDEPTH)
            self.x = I.shape[1]
            self.y = I.shape[0]
            self.dtype = I.dtype
            
        elif (file_extension.lower() == '.tif' or file_extension.lower() == '.tiff'): # Works with single-series tif & mutli-series ome.tif
            from skimage.external.tifffile import TiffFile
            self.filetype = 'tif'
            self.filehandle = TiffFile(filename)
            self.positions = len(self.filehandle.series)
            s = self.filehandle.series[0] # Here I'm going to assume all series have the same format
            self.timepoints = s.shape[s.axes.find('T')]
            self.channels = s.shape[s.axes.find('C')]
            self.x = s.shape[s.axes.find('X')]
            self.y = s.shape[s.axes.find('Y')]
            self.dtype = s.pages[0].asarray().dtype
                
    def close(self):
        if self.use_bioformats:
            self.filehandle.close()
            import javabridge
            javabridge.kill_vm()
        elif self.filetype == 'tif': # Nothing to do if sequence directory
            self.filehandle.close()
            
    def getfilenamefromprototype(self,position,channel,frame):
        '''
        Generate full filename for specific frame based on file path, 
        prototype, fileorder, and filenamesindexing

        Parameters
        ----------
        position : int
            Position/series index (0-based indexing).
        channel : int
            Imaging channel index (0-based indexing).
        frame : int
            Frame/timepoint index (0-based indexing).

        Returns
        -------
        string
            Filename.

        '''
        filenumbers = []
        
        for i in self.fileorder:
            if i=='p':
                filenumbers.append(position+self.filenamesindexing)
            if i=='c':
                filenumbers.append(channel+self.filenamesindexing)
            if i=='t':
                filenumbers.append(frame+self.filenamesindexing)
        return os.path.join(self.filehandle,self.prototype % tuple(filenumbers))
        
        
        
    def getframes(self, positions=None, channels=None, frames=None, 
                  squeeze_dimensions=True, 
                  resize=None, 
                  rescale=None,
                  globalrescale=None,
                  rotate=None):
        '''
        Get frames from experiment.

        Parameters
        ----------
        positions : None, int, tuple/list of ints, optional
            The frames from the position index or indexes passed as an integer 
            or a tuple/list will be returned. If None is passed, all positions 
            are returned. 
            The default is None.
        channels : None, int, tuple/list of ints, str, tuple/list of str, optional
            The frames from the channel index or indexes passed as an integer 
            or a tuple/list will be returned. If the channel names have been 
            defined, the channel(s) can be passed as a string or tuple/list of 
            strings. If an empty list is passed, None is returned. If None is 
            passed, all channels are returned.
            The default is None.
        frames : None, int, tuple/list of ints, optional
            The frame index or indexes passed as an integer or a tuple/list 
            will be returned. If None is passed, all frames are returned. 
            The default is None.
        squeeze_dimensions : bool, optional
            If True, the numpy squeeze function is applied to the output array,
            removing all singleton dimensions.
            The default is True.
        resize : None or tuple/list of 2 ints, optional
            Dimensions to resize the frames. If None, no resizing is performed.
            The default is None.
        rescale : None or tuple/list of 2 int/floats, optional
            Rescale all values in each frame to be within the given range.
            The default is None.
        globalrescale : None or tuple/list of 2 int/floats, optional
            Rescale all values in all frames to be within the given range.
            The default is None.
        rotate : None or float, optional
            Rotation to apply to the image (in degrees).
            The default is None.

        Raises
        ------
        ValueError
            If channel names are not correct.

        Returns
        -------
        Numpy Array
            Concatenated frames as requested by the different input options.
            If squeeze_dimensions=False, the array is 5-dimensional, with the
            dimensions order being: Position, Time, Channel, Y, X

        '''
        
        # Handle options:
        if frames is None: # No frames specfied: load all
            frames = list(range(self.timepoints))
        elif type(frames) is not list and type(frames) is not tuple:
            frames = [frames]
            
        if positions is None: # No positions specified: load all
            positions = list(range(self.positions))
        elif type(positions) is not list and type(positions) is not tuple:
            positions = [positions]
          
        if channels is None: # No channel specified: load all
            channels = list(range(self.channels))
        elif type(channels) is str: # refer to 1 channel by name (eg 'gfp')
            if self.channelnames is None:
                raise ValueError('Set channel names first')
            if channels in self.channelnames:
                channels = [self.channelnames.index(channels)]
            else:
                raise ValueError(channels+' is not a valid channel name.')
        elif (type(channels) is list or type(channels) is tuple):
            # If list of ints, nothing to do
            if len(channels)==0: # Empty list of channels: Return None
                return None
            elif all([type(c) is str for c in channels]): # refer to channels by list/tuple of names:
                if self.channelnames is None:
                    raise ValueError('Set channel names first')
                for i, c in enumerate(channels):
                    if c in self.channelnames:
                        channels[i] = self.channelnames.index(c)
                    else:
                        raise ValueError(c+' is not a valid channel name.')   
        elif type(channels) is not list and type(channels) is not tuple:
            channels = [channels]
            
        # Allocate memory:
        if rescale is None:
            dt = self.dtype
        else:
            dt = np.float32
        if resize is None:
            output = np.empty([len(positions),len(frames),len(channels),self.y,self.x],dtype=dt)
        else:
            output = np.empty([len(positions),len(frames),len(channels),resize[0],resize[1]],dtype=dt)
        
        # Load images:
        for p, pos in enumerate(positions):
            for c, cha in enumerate(channels):
                for f, fra in enumerate(frames):
                    # Read frame:
                    if self.use_bioformats:
                        frame = self.filehandle.read(series=pos,c=cha,t=fra,rescale=False)
                    elif self.filetype == 'dir':
                        frame = cv2.imread(self.getfilenamefromprototype(pos, cha, fra),cv2.IMREAD_ANYDEPTH)
                    elif self.filetype == 'tif':
                        frame = self.filehandle.series[pos].\
                                                    pages[fra*self.channels+cha].\
                                                    asarray()
                    # Optionally resize and rescale:
                    if rotate is not None:
                        M = cv2.getRotationMatrix2D((self.x/2,self.y/2), rotate, 1)
                        frame = cv2.warpAffine(frame,M,(self.x,self.y),borderMode=cv2.BORDER_REPLICATE)
                    if resize is not None:
                        frame = cv2.resize(frame, resize[::-1]) #cv2 inverts shape
                    if rescale is not None:
                        frame = rangescale(frame,rescale)
                    # Add to output array:
                    output[p,f,c,:,:] = frame
                    
        # Rescale all images:
        if globalrescale is not None:
            output = rangescale(output,globalrescale)

        # Return:
        return np.squeeze(output) if squeeze_dimensions else output
    


import importlib
import itertools
import warnings

import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
from skimage.exposure import rescale_intensity
from skimage.transform import rescale
from skimage.transform import rotate

div_odd = lambda n: (n // 2, n // 2 + 1)
perc_diff = lambda a, b: abs(a - b) / ((a + b) / 2)

if importlib.util.find_spec("cupy") is None:
    from scipy.signal import convolve2d as cuconvolve

    warnings.warn("Could not load CuPy for SyMBac, are you using a GPU? Defaulting to CPU convolution.")


    def convolve_rescale(image, kernel, rescale_factor, rescale_int):
        """
        Convolves an image with a kernel, and rescales it to the correct size. 

        Parameters
        ----------
        image : 2D numpy array
            The image
        kernel : 2D numpy array
            The kernel
        rescale_factor : int
            Typicall 1/resize_amount. So 1/3 will scale the image down by a factor of 3. We do this because we render the image and kernel at high resolution, so that we can do the convolution at high resolution.
        rescale_int : bool
            If True, rescale the intensities between 0 and 1 and return a float32 numpy array of the convolved downscaled image.

        Returns
        -------
        outupt : 2D numpy array
            The output of the convolution rescale operation
        """

        output = cuconvolve(image, kernel, mode="same")
        # output = output.get()
        output = rescale(output, rescale_factor, anti_aliasing=False)

        if rescale_int:
            output = rescale_intensity(output.astype(np.float32), out_range=(0, 1))
        return output
else:
    import cupy as cp
    from cupyx.scipy.ndimage import convolve as cuconvolve


    def convolve_rescale(image, kernel, rescale_factor, rescale_int):
        """
        Convolves an image with a kernel, and rescales it to the correct size. 

        Parameters
        ----------
        image : 2D numpy array
            The image
        kernel : 2D numpy array
            The kernel
        rescale_factor : int
            Typicall 1/resize_amount. So 1/3 will scale the image down by a factor of 3. We do this because we render the image and kernel at high resolution, so that we can do the convolution at high resolution.
        rescale_int : bool
            If True, rescale the intensities between 0 and 1 and return a float32 numpy array of the convolved downscaled image.

        Returns
        -------
        outupt : 2D numpy array
            The output of the convolution rescale operation
        """

        output = cuconvolve(cp.array(image), cp.array(kernel))
        output = output.get()
        output = rescale(output, rescale_factor, anti_aliasing=False)

        if rescale_int:
            output = rescale_intensity(output.astype(np.float32), out_range=(0, 1))
        return output


def generate_curve_props(cell_timeseries):
    """
    Generates individual cell curvature properties. 3 parameters for each cell, which are passed to a cosine function to modulate the cell's curvature. 
    
    Paramters
    ---------
    cell_timeseries : list
        The output of run_simulation()
    
    Returns
    -------
    outupt : A numpy array of unique curvature properties for each cell in the simulation
    """

    # Get unique cell IDs
    IDs = []
    for cell_list in cell_timeseries:
        for cell in cell_list:
            IDs.append(cell.ID)
    IDs = np.array(IDs)
    unique_IDs = np.unique(IDs)
    # For each cell, assign random curvature properties
    ID_props = []
    for ID in unique_IDs:
        freq_modif = (np.random.uniform(0.9, 1.1))  # Choose one per cell
        amp_modif = (np.random.uniform(0.9, 1.1))  # Choose one per cell
        phase_modif = np.random.uniform(-1, 1)  # Choose one per cell
        ID_props.append([int(ID), freq_modif, amp_modif, phase_modif])
    ID_propps = np.array(ID_props)
    ID_propps[:, 0] = ID_propps[:, 0].astype(int)
    return np.array(ID_props)


def gen_cell_props_for_draw(cell_timeseries_lists, ID_props):
    """
    Parameters
    ----------
    cell_timeseries_lists : list
        A list (single frame) from cell_timeseries, the output from run_simulation. E.g: cell_timeseries[x]
    ID_props : list
        A list of properties for each cell in that frame, the output of generate_curve_props()
    
    Returns
    -------
    cell_properties : list
        The final property list used to actually draw a scene of cells. The input to draw_scene
    """
    
    cell_properties = []
    for cell in cell_timeseries_lists:
        body, shape = (cell.body, cell.shape)
        vertices = []
        for v in shape.get_vertices():
            x, y = v.rotated(shape.body.angle) + shape.body.position  # .rotated(self.shape.body.angle)
            vertices.append((x, y))
        vertices = np.array(vertices)

        centroid = get_centroid(vertices)
        farthest_vertices = find_farthest_vertices(vertices)
        length = get_distance(farthest_vertices[0], farthest_vertices[1])
        width = cell.width
        # angle = np.arctan(vertices_slope(farthest_vertices[0], farthest_vertices[1]))
        angle = np.arctan2((farthest_vertices[0] - farthest_vertices[1])[1],
                           (farthest_vertices[0] - farthest_vertices[1])[0])
        angle = np.rad2deg(angle) + 90

        ID, freq_modif, amp_modif, phase_modif = ID_props[ID_props[:, 0] == cell.ID][0]
        phase_mult = 20
        cell_properties.append([length, width, angle, centroid, freq_modif, amp_modif, phase_modif, phase_mult])
    return cell_properties


def raster_cell(length, width):
    L = int(np.rint(length))
    W = int(np.rint(width))
    new_cell = np.zeros((L, W))
    R = (W - 1) / 2

    x_cyl = np.arange(0, 2 * R + 1, 1)
    I_cyl = np.sqrt(R ** 2 - (x_cyl - R) ** 2)
    L_cyl = L - W
    new_cell[int(W / 2):-int(W / 2), :] = I_cyl

    x_sphere = np.arange(0, int(W / 2), 1)
    sphere_Rs = np.sqrt((R) ** 2 - (x_sphere - R) ** 2)
    sphere_Rs = np.rint(sphere_Rs).astype(int)

    for c in range(len(sphere_Rs)):
        R_ = sphere_Rs[c]
        x_cyl = np.arange(0, R_, 1)
        I_cyl = np.sqrt(R_ ** 2 - (x_cyl - R_) ** 2)
        new_cell[c, int(W / 2) - sphere_Rs[c]:int(W / 2) + sphere_Rs[c]] = np.concatenate((I_cyl, I_cyl[::-1]))
        new_cell[L - c - 1, int(W / 2) - sphere_Rs[c]:int(W / 2) + sphere_Rs[c]] = np.concatenate((I_cyl, I_cyl[::-1]))
    new_cell = new_cell.astype(int)
    return new_cell


def get_distance(vertex1, vertex2):
    """
    Get euclidian distance between two sets of vertices.

    Parameters
    ----------
    vertex1 : 2-tuple
        x,y coordinates of a vertex
    vertex2 : 2-tuple
        x,y coordinates of a vertex

    Returns
    -------
    float : absolute distance between two points
    """
    return abs(np.sqrt((vertex1[0] - vertex2[0]) ** 2 + (vertex1[1] - vertex2[1]) ** 2))


def find_farthest_vertices(vertex_list):
    """Given a list of vertices, find the pair of vertices which are farthest from each other

    Parameters
    ----------
    vertex_list : list(2-tuple, 2-tuple ... )
        List of pairs of vertices [(x,y), (x,y), ...]

    Returns
    -------
    array(2-tuple, 2-tuple)
        The two vertices maximally far apart
    """
    vertex_combs = list(itertools.combinations(vertex_list, 2))
    distance = 0
    farthest_vertices = 0
    for vertex_comb in vertex_combs:
        distance_ = get_distance(vertex_comb[0], vertex_comb[1])
        if distance_ > distance:
            distance = distance_
            farthest_vertices = vertex_comb
    return np.array(farthest_vertices)


def get_midpoint(vertex1, vertex2):
    """
    Get the midpoint between two vertices
    """
    x_mid = (vertex1[0] + vertex2[0]) / 2
    y_mid = (vertex1[1] + vertex2[1]) / 2
    return np.array([x_mid, y_mid])


def vertices_slope(vertex1, vertex2):
    """
    Get the slope between two vertices
    """
    return (vertex1[1] - vertex2[1]) / (vertex1[0] - vertex2[0])


def midpoint_intercept(vertex1, vertex2):
    """
    Get the y-intercept of the line connecting two vertices
    """
    midpoint = get_midpoint(vertex1, vertex2)
    slope = vertices_slope(vertex1, vertex2)
    intercept = midpoint[1] - (slope * midpoint[0])
    return intercept


def get_centroid(vertices):
    """Return the centroid of a list of vertices 
    
    Keyword arguments:
    vertices -- A list of tuples containing x,y coordinates.

    """
    return np.sum(vertices, axis=0) / len(vertices)


def place_cell(length, width, angle, position, space):
    """Creates a cell and places it in the pymunk space

    Parameters
    ----------
    length : float
        length of the cell
    width : float
        width of the cell
    angle : float
        rotation of the cell in radians counterclockwise
    position : tuple
        x,y coordinates of the cell centroid
    space : pymunk.space.Space
        Pymunk space to place the cell in

    Returns
    -------
    nothing, updates space

    """
    angle = np.rad2deg(angle)
    x, y = np.array(position).astype(int)
    OPL_cell = raster_cell(length=length, width=width)
    rotated_OPL_cell = rotate(OPL_cell, angle, resize=True, clip=False, preserve_range=True)
    cell_y, cell_x = (np.array(rotated_OPL_cell.shape) / 2).astype(int)
    offset_y = rotated_OPL_cell.shape[0] - space[y - cell_y:y + cell_y, x - cell_x:x + cell_x].shape[0]
    offset_x = rotated_OPL_cell.shape[1] - space[y - cell_y:y + cell_y, x - cell_x:x + cell_x].shape[1]
    space[y - cell_y + 100:y + cell_y + offset_y + 100,
    x - cell_x + 100:x + cell_x + offset_x + 100] += rotated_OPL_cell


def transform_func(amp_modif, freq_modif, phase_modif):
    def perm_transform_func(x, amp_mult, freq_mult, phase_mult):
        return (amp_mult * amp_modif * np.cos(
            (x / (freq_mult * freq_modif) - phase_mult * phase_modif) * np.pi)).astype(int)

    return perm_transform_func


def scene_plotter(scene_array, output_dir, name, a, matplotlib_draw):
    if matplotlib_draw == True:
        plt.figure(figsize=(3, 10))
        plt.imshow(scene_array)
        plt.tight_layout()
        plt.savefig(output_dir + "/{}_{}.png".format(name, str(a).zfill(3)))
        plt.clf()
        plt.close('all')
    else:
        im = Image.fromarray(scene_array.astype(np.uint8))
        im.save(output_dir + "/{}_{}.tif".format(name, str(a).zfill(3)))


def make_images_same_shape(real_image, synthetic_image, rescale_int=True):
    """ Makes a synthetic image the same shape as the real image"""

    assert real_image.shape[0] < synthetic_image.shape[
        0], "Real image has a higher diemsion on axis 0, increase y_border_expansion_coefficient"
    assert real_image.shape[1] < synthetic_image.shape[
        1], "Real image has a higher diemsion on axis 1, increase x_border_expansion_coefficient"

    x_diff = synthetic_image.shape[1] - real_image.shape[1]
    remove_from_left, remove_from_right = div_odd(x_diff)
    y_diff = synthetic_image.shape[0] - real_image.shape[0]
    if real_image.shape[1] % 2 == 0:
        if synthetic_image.shape[1] % 2 == 0:
            if y_diff > 0:
                synthetic_image = synthetic_image[y_diff:, remove_from_left - 1:-remove_from_right]
            else:
                synthetic_image = synthetic_image[:, remove_from_left:-remove_from_right]
                real_image = real_image[abs(y_diff):, :]
        elif synthetic_image.shape[1] % 2 == 1:
            if y_diff > 0:
                synthetic_image = synthetic_image[y_diff:, remove_from_left:-remove_from_right]
            else:
                synthetic_image = synthetic_image[:, remove_from_left:-remove_from_right]
                real_image = real_image[abs(y_diff):, :]
    elif real_image.shape[1] % 2 == 1:
        if synthetic_image.shape[1] % 2 == 0:
            if y_diff > 0:
                synthetic_image = synthetic_image[y_diff:, remove_from_left:-remove_from_right]
            else:
                synthetic_image = synthetic_image[:, remove_from_left:-remove_from_right]
                real_image = real_image[abs(y_diff):, :]
        elif synthetic_image.shape[1] % 2 == 1:
            if y_diff > 0:
                synthetic_image = synthetic_image[y_diff:, remove_from_left - 1:-remove_from_right]
            else:
                synthetic_image = synthetic_image[:, remove_from_left:-remove_from_right]
                real_image = real_image[abs(y_diff):, :]

    if rescale_int:
        real_image = rescale_intensity(real_image.astype(np.float32), out_range=(0, 1))
        synthetic_image = rescale_intensity(synthetic_image.astype(np.float32), out_range=(0, 1))
    return real_image, synthetic_image


def get_space_size(cell_timeseries_properties):
    """Iterates through the simulation timeseries properties, 
    finds the extreme cell positions and retrieves the required 
    image size to fit all cells into"""
    max_x, max_y = 0, 0
    for timepoint in cell_timeseries_properties:
        for cell in timepoint:
            x_, y_ = np.ceil(cell[3]).astype(int)
            length_ = np.ceil(cell[0]).astype(int)
            width_ = np.ceil(cell[1]).astype(int)
            max_y_ = y_ + length_
            max_x_ = x_ + width_
            if max_x_ > max_x:
                max_x = max_x_
            if max_y_ > max_y:
                max_y = max_y_
    return (int(1.2 * max_y), int(1.5 * max_x))



'''
This script runs the segmentation U-Net on data that has been pre-processed to
crop out chamber images. See delta-interfacing/preprocessing.m or pipeline.py.

The images are processed by batches of 4096 to prevent memory issues.

@author: jblugagne
'''
from data import saveResult_seg, predictGenerator_seg, postprocess
from model import unet_seg
from os import listdir

# Files:
DeLTA_data = 'C:/DeepLearning/DeLTA_data/'
inputs_folder = DeLTA_data + 'mother_machine/evaluation/preprocessed/img/'
outputs_folder = DeLTA_data + 'mother_machine/evaluation/seg_output/'
model_file = DeLTA_data + 'mother_machine/models/unet_moma_seg_multisets.hdf5'
unprocessed = listdir(inputs_folder)

# Parameters:
target_size = (256, 32)
input_size = target_size + (1,)
process_size = 4096

# Load up model:
model = unet_seg(input_size = input_size)
model.load_weights(model_file)

# Process
while(unprocessed):
    # Pop out filenames
    ps = min(process_size,len(unprocessed))
    to_process = unprocessed[0:ps]
    del unprocessed[0:ps]
    
    # Predict:
    predGene = predictGenerator_seg(inputs_folder, files_list = to_process, target_size = target_size)
    results = model.predict_generator(predGene,len(to_process),verbose=1)
    
    # Post process results:
    results[:,:,:,0] = postprocess(results[:,:,:,0])
    
    # Save to disk:
    saveResult_seg(outputs_folder,results, files_list = to_process)


import pymunk
import numpy as np
def segment_creator(local_xy1, local_xy2,global_xy,thickness):
    segment_body = pymunk.Body(mass=2, moment=3,body_type=pymunk.Body.STATIC)
    segment_shape = pymunk.Segment(segment_body, local_xy1,local_xy2,thickness)
    segment_body.position = global_xy
    segment_shape.friction = 0
    return segment_body, segment_shape

def trench_creator(size,trench_length, global_xy, space):
    size = int(np.ceil(size/1.5))
    segments = []
    for x in range(size):
        segment = segment_creator((x,0),(0,size-x),global_xy,1)
        segments.append(segment)

    for x in range(size):
        segment = segment_creator((size-x,0),(size,size-x),(global_xy[0]+size/2, global_xy[1]),1)
        segments.append(segment)
    for z in segments:
        for s in z:
            space.add(s)

    left_wall = segment_creator((0,0),(0,trench_length),global_xy,1)
    right_wall = segment_creator((size,0),(size,trench_length),(global_xy[0]+size/2, global_xy[1]),1)
    barrier_thickness = 10
    left_barrier = segment_creator((0,0),(0,trench_length),(global_xy[0]-barrier_thickness, global_xy[1]),barrier_thickness)
    right_barrier = segment_creator((size,0),(size,trench_length),(global_xy[0]+size/2+barrier_thickness, global_xy[1]),barrier_thickness)
    walls = [left_wall, right_wall, left_barrier, right_barrier]
    for z in walls:
        for s in z:
            space.add(s)


'''
This script trains the tracking U-Net.

@author: jblugagne
'''
from model import unet_track
from data import trainGenerator_track
from tensorflow.keras.callbacks import ModelCheckpoint, EarlyStopping
import os
import config as cfg

# Parameters:
class_weights = (1,1,1) # same weight for all classes
batch_size = 5
epochs = 400
steps_per_epoch = 250
patience = 30

#Data generator parameters:
data_gen_args = dict(rotation = 1,
                     shiftX=.05,
                     shiftY=.05,
                     zoom=.15,
                     horizontal_flip=True,
                     histogram_voodoo=True,
                     illumination_voodoo=True)

# Generator init:
myGene = trainGenerator_track(batch_size,
                              os.path.join(cfg.training_set_track,'img'),
                              os.path.join(cfg.training_set_track,'seg'),
                              os.path.join(cfg.training_set_track,'previmg'),
                              os.path.join(cfg.training_set_track,'segall'),
                              os.path.join(cfg.training_set_track,'mother'),
                              os.path.join(cfg.training_set_track,'daughter'),
                              data_gen_args,
                              target_size = cfg.target_size_seg)

# Define model:
model = unet_track(input_size = cfg.target_size_seg+(4,),
                   class_weights = class_weights)
model.summary()

# Callbacks:
model_checkpoint = ModelCheckpoint(cfg.model_file_track,
                                   monitor='loss',
                                   verbose=1,
                                   save_best_only=True)
early_stopping = EarlyStopping(monitor='loss',
                               mode='min',
                               verbose=1,
                               patience=patience)

# Train:
history = model.fit_generator(myGene,
                              steps_per_epoch=steps_per_epoch,
                              epochs=epochs,
                              callbacks=[model_checkpoint, early_stopping])


'''
This file contains function definitions for data manipulations and input/output
operations.

@author: jblugagne
'''
from __future__ import print_function
import numpy as np 
import os, glob, re, random, warnings, copy, importlib
import skimage.io as io
import tifffile
import skimage.transform as trans
from skimage.measure import label
from skimage.morphology import square, binary_opening, remove_small_objects
from scipy import interpolate

# Try to import elastic deformations, issue warning if not found:
if importlib.util.find_spec("elasticdeform") is None:
    warnings.warn("Could not load elastic deformations module.")
else:
    import elasticdeform
        



#%% UTILITIES:

def binarizerange(i):
    '''
    This function will binarize a numpy array by thresholding it in the middle
    of its range

    Parameters
    ----------
    i : 2D numpy array
        Input array/image.

    Returns
    -------
    newi : 2D numpy array
        Binarized image.

    '''
    newi = i
    newi[i > (np.amin(i)+np.amax(i))/2] = 1
    newi[i <= (np.amin(i)+np.amax(i))/2] = 0
    return newi

def readreshape(filename, target_size = (256,32), binarize = False, order = 1, rangescale = True):
    '''
    Read image from disk and format it

    Parameters
    ----------
    filename : string
        Path to file. Only PNG, JPG or single-page TIFF files accepted
    target_size : tupe of int, optional
        Size to reshape the image. 
        The default is (256,32).
    binarize : bool, optional
        Use the binarizerange() function on the image.
        The default is False.
    order : int, optional
        interpolation order (see skimage.transform.warp doc). 
        0 is nearest neighbor
        1 is bilinear
        The default is 1.
    rangescale : bool, optional
        Scale array image values to 0-1 if True. 
        The default is True.

    Raises
    ------
    ValueError
        Raised if image file is not a PNG, JPEG, or TIFF file.

    Returns
    -------
    i : numpy 2d array of floats
        Loaded array.

    '''
    fext = os.path.splitext(filename)[1].lower()
    if fext in ('.png', '.jpg', '.jpeg'):
        i = io.imread(filename,as_gray = True)
    elif fext in ('.tif', '.tiff'):
        i = tifffile.imread(filename)
    else:
        raise ValueError('Only PNG, JPG or single-page TIF files accepted')
        
    i = trans.resize(i, target_size, anti_aliasing=True, order=order)
    if binarize:
        i = binarizerange(i)
    i = np.reshape(i,i.shape + (1,) )
    if rangescale:
        i = (i-np.min(i))/np.ptp(i)
    return i

def postprocess(images,square_size=5,min_size=None):
    '''
    A generic binary image cleaning function based on mathematical morphology.

    Parameters
    ----------
    images : 2D or 3D numpy array
        Input image or images stacked along axis=0.
    square_size : int, optional
        Size of the square structuring element to use for morphological opening
        The default is 5.
    min_size : int or None, optional
        Remove objects smaller than this minimum area value. If None, the 
        operation is not performed.
        The default is None.

    Returns
    -------
    images : 2D numpy array
        Cleaned, binarized images. Note that the dimensions are squeezed before 
        return (see numpy.squeeze doc)

    '''
    
    # Expand dims if 2D:
    if images.ndim == 2:
        images = np.expand_dims(images,axis=0)
    
    # Generate struturing element:
    selem = square(square_size)
    
    # Go through stack:
    for index, I in enumerate(images):
        I = binarizerange(I)
        I = binary_opening(I,selem=selem)
        if min_size is not None:
            I = remove_small_objects(I,min_size=min_size)
        images[index,:,:] = I
        
    return np.squeeze(images)
        
# def weightmap(mask, classweights, params):
#     '''
#     This function computes the weight map as described in the original U-Net
#     paper to force the model to learn borders 
#     (Not used, too slow. We preprocess the weight maps in matlab)
#     '''

#     lblimg,lblnb = label(mask[:,:,0],connectivity=1,return_num=True)
#     distance_array = float('inf')*np.ones(mask.shape[0:2] + (max(lblnb,2),))
#     for i in range(0,lblnb):
#         [_,distance_array[:,:,i]] = medial_axis((lblimg==i+1)==0,return_distance=True)
#     distance_array = np.sort(distance_array,axis=-1)
#     weightmap = params["w0"]*np.exp(-np.square(np.sum(distance_array[:,:,0:2],-1))/(2*params["sigma"]^2))
#     weightmap = np.multiply(weightmap,mask[:,:,0]==0)
#     weightmap = np.add(weightmap,(mask[:,:,0]==0)*classweights[0])
#     weightmap = np.add(weightmap,(mask[:,:,0]==1)*classweights[1])
#     weightmap = np.reshape(weightmap, weightmap.shape + (1,))
    
#     return weightmap



#%% DATA AUGMENTATION
def data_augmentation(images_input, aug_par, order=0):
    '''
    Data augmentation function

    Parameters
    ----------
    images_input : list of 2D numpy arrays of floats
        Images to apply augmentation operations to.
    aug_par : dict
        Augmentation operations parameters. Accepted key-value pairs:
            illumination_voodoo: bool. Whether to apply the illumination 
                voodoo operation.
            histogram_voodoo: bool. Whether to apply the histogram voodoo 
                operation.
            elastic_deformation: dict. If key exists, the elastic deformation
                operation is applied. The parameters are given as key-value 
                pairs. sigma values are given under the sigma key, deformation
                points are given under the points key. See elasticdeform doc.
            gaussian_noise: float. Apply gaussian noise to the image. The 
                sigma value of the gaussian noise is uniformly sampled between
                0 and +gaussian_noise.
            horizontal_flip: bool. Whether to flip the images horizontally. 
                Input images have a 50% chance of being flipped
            vertical_flip: bool. Whether to flip the images vertically. 
                Input images have a 50% chance of being flipped
            rotations90d: bool. Whether to randomly rotate the images in 90°
                increments. Each 90° rotation has a 25% chance of happening
            rotation: int/float. Range of random rotation to apply. The angle 
                is uniformly sampled in the range [-rotation, +rotation]
            zoom: float. Range of random "zoom" to apply. The image
                is randomly zoomed by a factor that is sampled from an 
                exponential distribution with a lamba of 3/zoom. The random
                factor is clipped at +zoom.
            shiftX: int/float. The range of random shifts to apply along X. A
                uniformly sampled shift between [-shiftX, +shiftX] is applied
            shiftY: int/float. The range of random shifts to apply along Y. A
                uniformly sampled shift between [-shiftY, +shiftY] is applied
            Note that the same operations are applied to all inputs.
    order : int or list/tuple of ints, optional
        Interpolation order to use for each image in the input stack. If order
        is a scalar, the same order is applied to all images. If a list of
        orders is provided, each image in the stack will have its own operaiton
        order. See skimage.transform.wrap doc.
        Note that the histogram voodoo operation is only applied to images with
        a non-zero order.
        The default is 0.

    Returns
    -------
    output : list of 2D numpy arrays of floats
        Augmented images array.

    '''
    
    #processing inputs / initializing variables::
    output = list(images_input)
    if np.isscalar(order):
        orderlist = [order] * len(images_input)
    else:
        orderlist = list(order)
    
    # Apply augmentation operations:
    
    if "illumination_voodoo" in aug_par:
        if aug_par["illumination_voodoo"]:
            for index, item in enumerate(output):
                if order[index] > 0: # Not super elegant, but tells me if binary or grayscale image
                    output[index] = illumination_voodoo(item)
                    
    #TODO: Illumination voodoo 2D
                    
    if "histogram_voodoo" in aug_par:
        if aug_par["histogram_voodoo"]:
            for index, item in enumerate(output):
                if order[index] > 0: # Not super elegant, but tells me if binary or grayscale image
                    output[index] = histogram_voodoo(item)
                    
    
    if 'gaussian_noise' in aug_par:
        if aug_par['gaussian_noise']:
            sigma = np.random.rand()*aug_par['gaussian_noise']
            for index, item in enumerate(output):
                if order[index] > 0: # Not super elegant, but tells me if binary or grayscale image
                    item = item + np.random.normal(0,sigma,item.shape) # Add Gaussian noise
                    output[index] = (item-np.min(item))/np.ptp(item) # Rescale to 0-1
                
    if "elastic_deformation" in aug_par:
        output = elasticdeform.deform_random_grid(output,
                                                  sigma=aug_par["elastic_deformation"]["sigma"],
                                                  points=aug_par["elastic_deformation"]["points"],
                                                  order=[i*3 for i in orderlist], # Using bicubic interpolation instead of bilinear here
                                                  mode='nearest',
                                                  axis=(0,1),
                                                  prefilter=False)
    
    if "horizontal_flip" in aug_par:
        if aug_par["horizontal_flip"]:
            if random.randint(0,1): #coin flip
                for index, item in enumerate(output):
                    output[index] = np.fliplr(item)
                    
    if "vertical_flip" in aug_par:
        if aug_par["vertical_flip"]:
            if random.randint(0,1): #coin flip
                for index, item in enumerate(output):
                    output[index] = np.flipud(item)
                    
    if "rotations_90d" in aug_par: # Only works with square images right now!
        if aug_par["rotations_90d"]:
            rot = random.randint(0,3)*90
            if rot>0: 
                for index, item in enumerate(output):
                    output[index] = trans.rotate(item,rot,mode='edge',order=orderlist[index])
    
    if "rotation" in aug_par:
        rot = random.uniform(-aug_par["rotation"],aug_par["rotation"])
        for index, item in enumerate(output):
            output[index] = trans.rotate(item,rot,mode='edge',order=orderlist[index])
    
    if "zoom" in aug_par:
        zoom = random.expovariate(3*1/aug_par["zoom"]) # I want most of them to not be too zoomed
        zoom = aug_par["zoom"] if zoom > aug_par["zoom"] else zoom
    else:
        zoom = 0
        
    if "shiftX" in aug_par:
        shiftX = random.uniform(-aug_par["shiftX"],aug_par["shiftX"])
    else:
        shiftX = 0
        
    if "shiftY" in aug_par:
        shiftY = random.uniform(-aug_par["shiftY"],aug_par["shiftY"])
    else:
        shiftY = 0
        
    if any(abs(x)>0 for x in [zoom,shiftX,shiftY]):
        for index, item in enumerate(output):
            output[index] = zoomshift(item,zoom+1,shiftX,shiftY, order=orderlist[index])
    
    return output

def zoomshift(I,zoomlevel,shiftX,shiftY, order=0):
    '''
    This function zooms and shifts images.

    Parameters
    ----------
    I : 2D numpy array
        input image.
    zoomlevel : float
        Additional zoom to apply to the image.
    shiftX : int
        X-axis shift to apply to the image, in pixels.
    shiftY : int
        Y-axis shift to apply to the image, in pixels.
    order : int, optional
        Interpolation order. The default is 0.

    Returns
    -------
    I : 2D numpy array
        Zoomed and shifted image of same size as input.

    '''
    
    oldshape = I.shape
    I = trans.rescale(I,zoomlevel,mode='edge',multichannel=False, order=order)
    shiftX = shiftX * I.shape[0]
    shiftY = shiftY * I.shape[1]
    I = shift(I,(shiftY, shiftX),order=order) # For some reason it looks like X & Y are inverted?
    i0 = (round(I.shape[0]/2 - oldshape[0]/2), round(I.shape[1]/2 - oldshape[1]/2))
    I = I[i0[0]:(i0[0]+oldshape[0]), i0[1]:(i0[1]+oldshape[1])]
    return I
    
def shift(image, vector, order=0):
    '''
    Image shifting function

    Parameters
    ----------
    image : 2D numpy array
        Input image.
    vector : tuple of ints
        Translation/shit vector.
    order : int, optional
        Interpolation order. The default is 0.

    Returns
    -------
    shifted : 2D numpy image
        Shifted image.

    '''
    transform = trans.AffineTransform(translation=vector)
    shifted = trans.warp(image, transform, mode='edge',order=order)

    return shifted

def histogram_voodoo(image,num_control_points=3):
    '''
    This function kindly provided by Daniel Eaton from the Paulsson lab.
    It performs an elastic deformation on the image histogram to simulate
    changes in illumination

    Parameters
    ----------
    image : 2D numpy array
        Input image.
    num_control_points : int, optional
        Number of inflection points to use on the histogram conversion curve. 
        The default is 3.

    Returns
    -------
    2D numpy array
        Modified image.

    '''
    control_points = np.linspace(0,1,num=num_control_points+2)
    sorted_points = copy.copy(control_points)
    random_points = np.random.uniform(low=0.1,high=0.9,size=num_control_points)
    sorted_points[1:-1] = np.sort(random_points)
    mapping = interpolate.PchipInterpolator(control_points, sorted_points)
    
    return mapping(image)

def illumination_voodoo(image,num_control_points=5):
    '''
    This function inspired by the one above.
    It simulates a variation in illumination along the length of the chamber

    Parameters
    ----------
    image : 2D numpy array
        Input image.
    num_control_points : int, optional
        Number of inflection points to use on the illumination multiplication
        curve. 
        The default is 5.

    Returns
    -------
    newimage : 2D numpy array
        Modified image.

    '''
    
    # Create a random curve along the length of the chamber:
    control_points = np.linspace(0,image.shape[0]-1,num=num_control_points)
    random_points = np.random.uniform(low=0.1,high=0.9,size=num_control_points)
    mapping = interpolate.PchipInterpolator(control_points, random_points)
    curve = mapping(np.linspace(0,image.shape[0]-1,image.shape[0]))
    # Apply this curve to the image intensity along the length of the chamebr:
    newimage = np.multiply(image,
                           np.reshape(
                                   np.tile(
                                           np.reshape(curve,curve.shape + (1,)), (1, image.shape[1])
                                           )
                                   ,image.shape
                                   )
                           )
    # Rescale values to original range:
    newimage = np.interp(newimage, (newimage.min(), newimage.max()), (image.min(), image.max()))
    
    return newimage



#%% SEGMENTATION FUNCTIONS:

def trainGenerator_seg(batch_size,
                    img_path, mask_path, weight_path,
                    target_size = (256,32),
                    augment_params = {},
                    preload = False,
                    seed = 1):
    '''
    Generator for training the segmentation U-Net.

    Parameters
    ----------
    batch_size : int
        Batch size, number of training samples to concatenate together.
    img_path : string
        Path to folder containing training input images.
    mask_path : string
        Path to folder containing training segmentation groundtruth.
    weight_path : string
        Path to folder containing weight map images.
    target_size : tuple of 2 ints, optional
        Input and output image size. 
        The default is (256,32).
    augment_params : dict, optional
        Data augmentation parameters. See data_augmentation() doc for more info
        The default is {}.
    preload : bool, optional
        Flag to load all training inputs in memory during intialization of the
        generator.
        The default is False.
    seed : int, optional
        Seed for numpy's random generator. see numpy.random.seed() doc
        The default is 1.

    Yields
    ------
    image_arr : 4D numpy array of floats
        Input images for the U-Net training routine. Dimensions of the tensor 
        are (batch_size, target_size[0], target_size[1], 1)
    mask_wei_arr : 4D numpy array of floats
        Output masks and weight maps for the U-Net training routine. Dimensions
        of the tensor are (batch_size, target_size[0], target_size[1], 2)

    '''
    
    preload_mask = []
    preload_img = []
    preload_weight = []
    
    # Get training image files list:
    image_name_arr =    glob.glob(os.path.join(img_path,"*.png")) +\
                        glob.glob(os.path.join(img_path,"*.tif")) +\
			glob.glob(os.path.join(img_path,"*.tiff"))
    
    # If preloading, load the images and compute weight maps:
    if preload:
        for filename in image_name_arr:
            preload_img.append(readreshape(filename, target_size = target_size, order = 1))
            preload_mask.append(readreshape(os.path.join(mask_path,os.path.basename(filename)), target_size = target_size, binarize = True, order = 0, rangescale = False))
            if weight_path is not None:
                preload_weight.append(readreshape(os.path.join(weight_path,os.path.basename(filename)), target_size = target_size, order = 0), rangescale = False)
    
    # Reset the pseudo-random generator:
    random.seed(a=seed)
    
    while True:
        # Reset image arrays:
        image_arr = []
        mask_arr = []
        weight_arr = []
        for _ in range(batch_size):
            # Pick random image index:
            index = random.randrange(0,len(image_name_arr))
            
            if preload:
                # Get from preloaded arrays:
                img = preload_img[index]
                mask = preload_mask[index]
                weight = preload_weight[index]
            else:
                # Read images:
                filename = image_name_arr[index]
                img = readreshape(filename, target_size = target_size, order = 1)
                mask = readreshape(os.path.join(mask_path,os.path.basename(filename)), target_size = target_size, binarize = True, order = 0, rangescale = False)
                if weight_path is not None:
                    weight = readreshape(os.path.join(weight_path,os.path.basename(filename)), target_size = target_size, order = 0, rangescale = False)
                else:
                    weight = []
                    

            # Data augmentation:
            if weight_path is not None:
                [img, mask, weight] = data_augmentation([img, mask, weight], augment_params, order=[1,0,0])
            else:
                [img, mask] = data_augmentation([img, mask], augment_params, order=[1,0])
            
            # Append to output list:
            image_arr.append(img)
            mask_arr.append(mask)
            weight_arr.append(weight)

        # Concatenate masks and weights: (gets unstacked by the loss function)
        image_arr = np.array(image_arr)
        if weight_path is not None:
            mask_wei_arr = np.concatenate((mask_arr,weight_arr),axis=-1)
        else:
            mask_wei_arr = np.array(mask_arr)
        yield (image_arr, mask_wei_arr)


def saveResult_seg(save_path,npyfile, files_list = [], multipage=False):
    '''
    Saves an array of segmentation output images to disk

    Parameters
    ----------
    save_path : string
        Path to save folder.
    npyfile : 3D or 4D numpy array
        Array of segmentation outputs to save to individual files. If 4D, only 
        the images from the first index of axis=3 will be saved.
    files_list : list of strings, optional
        Filenames to save the segmentation masks as. png, tif or jpg extensions
        work.
        The default is [].
    multipage : bool, optional
        Flag to save all output masks as a single, multi-page TIFF file. Note
        that if the file already exists, the masks will be appended to it.
        The default is False.

    Returns
    -------
    None.

    '''
    
    for i,item in enumerate(npyfile):
        if item.ndim == 3:
            img = item[:,:,0]
        else:
            img = item
        if multipage:
            filename = os.path.join(save_path,files_list[0])
            io.imsave(filename,(img*255).astype(np.uint8),plugin='tifffile',append=True)
        else:
            if files_list:
                filename = os.path.join(save_path,files_list[i])
            else:
                filename = os.path.join(save_path,"%d_predict.png"%i)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                io.imsave(filename,(img*255).astype(np.uint8))


def predictGenerator_seg(files_path, files_list = [], target_size = (256,32)):
    '''
    Get a generator for predicting segmentation on new image files 
    once the segmentation U-Net has been trained.

    Parameters
    ----------
    files_path : string
        Path to image files folder.
    files_list : list/tuple of strings, optional
        List of file names to read in the folder. If empty, all 
        files in the folder will be read.
        The default is [].
    target_size : tuple of 2 ints, optional
        Size for the images to be resized.
        The default is (256,32).

    Returns
    ------
    mygen : generator
        Generator that will yield single image files as 4D numpy arrays of
        size (1, target_size[0], target_size[1], 1).

    '''
    
    
    if not files_list:
        files_list = os.listdir(files_path)

        
    def generator(files_path, files_list, target_size):
        for index, fname in enumerate(files_list):
            img = readreshape(os.path.join(files_path,fname),
                              target_size=target_size,
                              order=1)
            img = np.reshape(img,(1,)+img.shape) # Tensorflow needs one extra single dimension (so that it is a 4D tensor)
            yield img
                
    mygen = generator(files_path, files_list, target_size)
    return mygen
    

        
        
        
#%% TRACKING FUNCTIONS
    
def trainGenerator_track(batch_size,
                    img_path, seg_path, previmg_path, segall_path,
                    mother_path, daughter_path,
                    augment_params = {},
                    target_size = (256,32),
                    seed = 1):
    '''
    Generator for training the tracking U-Net.

    Parameters
    ----------
    batch_size : int
        Batch size, number of training samples to concatenate together.
    img_path : string
        Path to folder containing training input images (current timepoint).
    seg_path : string
        Path to folder containing training 'seed' images, ie mask of 1 cell 
        in the previous image to track in the current image.
    previmg_path : string
        Path to folder containing training input images (previous timepoint).
    segall_path : string
        Path to folder containing training 'segall' images, ie mask of all 
        cells in the current image.
    mother_path : string
        Path to folder containing 'mother' tracking groundtruth, ie mask of 
        the seed cell tracked in the current frame. If a division happens
        between frames, the top cell is considered the 'mother' cell
    daughter_path : string
        Path to folder containing 'daughter' tracking groundtruth, ie mask of 
        the daughter of the seed cell in the current frame, if a division has 
        happened.
    augment_params : dict, optional
        Data augmentation parameters. See data_augmentation() doc for more info
        The default is {}.
    target_size : tuple of 2 ints, optional
        Input and output image size. 
        The default is (256,32).
    seed : int, optional
        Seed for numpy's random generator. see numpy.random.seed() doc
        The default is 1.

    Yields
    ------
    inputs_arr : 4D numpy array of floats
        Input images and masks for the U-Net training routine. Dimensions of 
        the tensor are (batch_size, target_size[0], target_size[1], 4)
    outputs_arr : 4D numpy array of floats
        Output masks for the U-Net training routine. Dimensions of the tensor 
        are (batch_size, target_size[0], target_size[1], 3). The third index
        of axis=3 contains 'background' masks, ie the part of the tracking 
        output groundtruth that is not part of the mother or daughter masks
    '''
    
    
    # Initialize variables and arrays:
    image_name_arr =    glob.glob(os.path.join(img_path,"*.png")) +\
                        glob.glob(os.path.join(img_path,"*.tif"))
    
    # Reset the pseudo-random generator:
    random.seed(a=seed)
    
    while True:
        # Reset image arrays:
        img_arr = []
        seg_arr = []
        previmg_arr = []
        segall_arr = []
        mother_arr = []
        daughter_arr = []
        bkgd_arr = []

        for _ in range(batch_size):
            # Pick random image file name:
            index = random.randrange(0,len(image_name_arr))
            filename = image_name_arr[index]
            
            # Read images:
            img = readreshape(filename, target_size = target_size, order = 1)
            seg = readreshape(os.path.join(seg_path,os.path.basename(filename)), target_size = target_size, binarize = True, order = 0, rangescale = False)
            previmg = readreshape(os.path.join(previmg_path,os.path.basename(filename)), target_size = target_size, order = 1)
            segall = readreshape(os.path.join(segall_path,os.path.basename(filename)), target_size = target_size, binarize = True, order = 0, rangescale = False)

            mother = readreshape(os.path.join(mother_path,os.path.basename(filename)), target_size = target_size, binarize = True, order = 0, rangescale = False)
            daughter = readreshape(os.path.join(daughter_path,os.path.basename(filename)), target_size = target_size, binarize = True, order = 0, rangescale = False)
            
            # Create the "background" image: (necessary for categorical cross-entropy)
            bkgd = 1 - (mother + daughter)
            
            # Data augmentation:
            [img, seg, previmg, segall, mother, daughter, bkgd] = data_augmentation([img, seg, previmg, segall, mother, daughter, bkgd],
                                augment_params, order=[1,0,1,0,0,0,0])
            
            # Append to arrays:
            seg_arr.append(seg)
            img_arr.append(img)
            previmg_arr.append(previmg)
            segall_arr.append(segall)
            mother_arr.append(mother)
            daughter_arr.append(daughter)
            bkgd_arr.append(bkgd)
            
        # Concatenare and yield inputs and output:
        inputs_arr = np.concatenate((img_arr,seg_arr,previmg_arr,segall_arr),axis=-1)
        outputs_arr = np.concatenate((mother_arr,daughter_arr,bkgd_arr),axis=-1)
        yield (inputs_arr, outputs_arr)
            
            
# def predictGenerator_track(test_path,files_list = [],target_size = (256,32)):
#     '''
#     This function is a generator for predicting tracking on new image files 
#     once the tracking U-Net has been trained.
    
#     Images are loaded from subfolders in the test_path folder (following the 
#     same nomenclature as in trainGenerator_track())
#     If a list of filenames files_list is specified, it will read only those 
#     images, otherwise it will go through the entire folder.
    
#     Images are resized to target_size upon loading.
#     '''
    
#     if not files_list:
#         files_list = os.listdir(test_path)
#     for index, item in enumerate(files_list):
#             img = readreshape(os.path.join(test_path + 'img/',os.path.basename(item)), target_size = target_size, order = 1)
#             seg = readreshape(os.path.join(test_path + 'seg/',os.path.basename(item)), target_size = target_size, binarize = True, order = 0, rangescale = False)
#             previmg = readreshape(os.path.join(test_path + 'previmg/',os.path.basename(item)), target_size = target_size, order = 1)
#             segall = readreshape(os.path.join(test_path + 'segall/',os.path.basename(item)), target_size = target_size, binarize = True, order = 0, rangescale = False)
            
#             inputs_arr = np.concatenate((img,seg,previmg,segall),axis=-1)
#             inputs_arr = np.reshape(inputs_arr,(1,)+inputs_arr.shape) # Tensorflow needs one extra single dimension (so that it is a 4D tensor)
            
#             yield inputs_arr


def predictCompilefromseg_track(img_path,seg_path, files_list = [],
                                  target_size = (256,32)):
    '''
    Compile an inputs array for tracking prediction with the tracking U-Net, 
    directly from U-Net segmentation masks saved to disk.

    Parameters
    ----------
    img_path : string
        Path to original single-chamber images folder. The filenames are 
        expected in the printf format Position%02d_Chamber%02d_Frame%03d.png
    seg_path : string
        Path to segmentation output masks folder. The filenames must be the 
        same as in the img_path folder.
    files_list : tuple/list of strings, optional
        List of filenames to compile in the img_path and seg_path folders. 
        If empty, all files in the folder will be read.
        The default is [].
    target_size : tuple of 2 ints, optional
        Input and output image size. 
        The default is (256,32).

    Returns
    -------
    inputs_arr : 4D numpy array of floats
        Input images and masks for the tracking U-Net training routine. 
        Dimensions of the tensor are (cells_to_track, target_size[0], 
        target_size[1], 4), with cells_to_track the number of segmented cells
        in all segmentation masks of the files_list.
    seg_name_list : list of strings
        Filenames to save the tracking outputs as. The printf format is 
        Position%02d_Chamber%02d_Frame%03d_Cell%02d.png, with the '_Cell%02d'
        string appended to signal which cell is being seeded/tracked (from top
        to bottom)
        

    '''
    
    img_arr = []
    seg_arr = []
    previmg_arr = []
    segall_arr = []
    seg_name_list = []
    
    if not files_list:
        files_list = os.listdir(img_path)
        
    for index, item in enumerate(files_list):
            filename = os.path.basename(item)
            # Get position, chamber & frame numbers:
            (pos, cha, fra) = list(map(int, re.findall("\d+",filename)))
            if fra > 1:
                prevframename = 'Position' + str(pos).zfill(2) + \
                                '_Chamber' + str(cha).zfill(2) + \
                                '_Frame' + str(fra-1).zfill(3) + '.png'
                img = readreshape(os.path.join(img_path,filename), target_size = target_size, order = 1)
                segall = readreshape(os.path.join(seg_path,filename), target_size = target_size, order = 0, binarize = True, rangescale = False)
                previmg = readreshape(os.path.join(img_path,prevframename), target_size = target_size, order = 1)
                prevsegall = readreshape(os.path.join(seg_path,prevframename), target_size = target_size, order = 0, binarize = True, rangescale = False)
                
                lblimg,lblnb = label(prevsegall[:,:,0],connectivity=1,return_num=True)
                
                for lbl in range(1,lblnb+1):
                    segfilename = 'Position' + str(pos).zfill(2) + \
                                '_Chamber' + str(cha).zfill(2) + \
                                '_Frame' + str(fra).zfill(3) + \
                                '_Cell' + str(lbl).zfill(2) + '.png'

                    seg = lblimg == lbl
                    seg = np.reshape(seg, seg.shape + (1,))
                    seg.astype(segall.dtype) # Output is boolean otherwise

                    seg_arr.append(seg)
                    img_arr.append(img)
                    previmg_arr.append(previmg)
                    segall_arr.append(segall)
                    seg_name_list.append(segfilename)
    
    inputs_arr = np.concatenate((img_arr,seg_arr,previmg_arr,segall_arr),axis=-1)
    return inputs_arr, seg_name_list


def estimateClassweights(gene, num_samples = 30):
    '''
    Estimate the class weights to use with the weighted 
    categorical cross-entropy based on the output of the trainGenerator_track
    output.

    Parameters
    ----------
    gene : generator
        Tracking U-Net training generator. (output of trainGenerator_track)
    num_samples : int, optional
        Number of batches to use for estimation. The default is 30.

    Returns
    -------
    class_weights : tuple of floats
        Relative weights of each class. Note that, if 0 elements of a certain 
        class are present in the samples, the weight for this class will be set 
        to 0.

    '''
    
    sample = next(gene)
    class_counts = [0] * sample[1].shape[-1] # List of 0s
    
    # Run through samples and classes/categories:
    for _ in range(num_samples):
        for i in range(sample[1].shape[-1]):
            class_counts[i] += np.mean(sample[1][..., i])
        sample = next(gene)
        
    # Warning! If 0 elements of a certain class are present in the samples, the
    # weight for this class will be set to 0. This is for the tracking case
    # (Where there are only empty daughter images in the training set)
    # Try changing the num_samples value if this is a problem
    class_weights = [(x/num_samples)**-1 if x!= 0 else 0 for x in class_counts] # Normalize by nb of samples and invert to get weigths, unless x == 0 to avoid Infinite weights or errors
    
    return class_weights


def saveResult_track(save_path,npyfile, files_list = None):
    '''
    Save tracking output masks to disk

    Parameters
    ----------
    save_path : string
        Folder to save images to.
    npyfile : 4D numpy array
        Array of tracking outputs to save to individual files.
    files_list : tuple/list of strings, optional
        Filenames to save the masks as. Note that the 'mother_' and 'daughter_'
        prefixes will be added to those names. If None, numbers will be used.
        The default is None.

    Returns
    -------
    None.

    '''
    
    mothers = npyfile[:,:,:,0]
    daughters = npyfile[:,:,:,1]
    for i,mother in enumerate(mothers):
        if files_list:
            filenameMo = os.path.join(save_path,'mother_' + files_list[i])
        else:
            filenameMo = os.path.join(save_path,"%mother_09d.png"%i)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            io.imsave(filenameMo,(mother*255).astype(np.uint8))
    for i,daughter in enumerate(daughters):
        if files_list:
            filenameDa = os.path.join(save_path,'daughter_' + files_list[i])
        else:
            filenameDa = os.path.join(save_path,"%daughter_09d.png"%i)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            io.imsave(filenameDa,(daughter*255).astype(np.uint8))



'''
This file contains model definitions and loss/metrics functions definitions

@author: jblugagne
'''
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras import backend as K
from tensorflow.python.ops import array_ops, math_ops
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Conv2D, MaxPooling2D, Dropout, UpSampling2D, Concatenate
from tensorflow.keras.optimizers import Adam



#%% Losses/metrics
def pixelwise_weighted_binary_crossentropy(y_true, y_pred):
    '''
    Pixel-wise weighted binary cross-entropy loss.
    The code is adapted from the Keras TF backend.
    (see their github)
    
    Parameters
    ----------
    y_true : Tensor
        Stack of groundtruth segmentation masks + weight maps.
    y_pred : Tensor
        Predicted segmentation masks.

    Returns
    -------
    Tensor
        Pixel-wise weight binary cross-entropy between inputs.

    '''
    
    try:
        # The weights are passed as part of the y_true tensor:
        [seg, weight] = tf.unstack(y_true, 2, axis=-1)

        seg = tf.expand_dims(seg, -1)
        weight = tf.expand_dims(weight, -1)
    except:
        pass

    epsilon = tf.convert_to_tensor(K.epsilon(), y_pred.dtype.base_dtype)
    y_pred = tf.clip_by_value(y_pred, epsilon, 1. - epsilon)
    y_pred = tf.math.log(y_pred / (1 - y_pred))

    zeros = array_ops.zeros_like(y_pred, dtype=y_pred.dtype)
    cond = (y_pred >= zeros)
    relu_logits = math_ops.select(cond, y_pred, zeros)
    neg_abs_logits = math_ops.select(cond, -y_pred, y_pred)
    entropy = math_ops.add(relu_logits - y_pred * seg, math_ops.log1p(math_ops.exp(neg_abs_logits)), name=None)
    
    # This is essentially the only part that is different from the Keras code:
    return K.mean(math_ops.multiply(weight, entropy), axis=-1)


def class_weighted_categorical_crossentropy(class_weights):
    '''
    Generate class-weighted categorical cross-entropy loss function.
    The code is adapted from the Keras TF backend.
    (see their github)
    
    Parameters
    ----------
    class_weights : tuple/list of floats
        Weights for each class/category.

    Returns
    -------
    function.
        Class-weighted categorical cross-entropy loss function.

    '''
    
    def loss_function(y_true, y_pred):
        
        # scale preds so that the class probas of each sample sum to 1
        y_pred /= tf.reduce_sum(y_pred, -1, True)
        # manual computation of crossentropy
        epsilon = tf.convert_to_tensor(K.epsilon(), y_pred.dtype.base_dtype)
        y_pred = tf.clip_by_value(y_pred, epsilon, 1. - epsilon)
        
        # Multiply each class by its weight:
        classes_list = tf.unstack(y_true * tf.math.log(y_pred), axis=-1)
        for i in range(len(classes_list)):
            classes_list[i] = tf.scalar_mul(class_weights[i], classes_list[i])
        
        # Return weighted sum:
        return - tf.reduce_sum(tf.stack(classes_list, axis=-1), -1)
    
    return loss_function


def unstack_acc(y_true, y_pred):
    '''
    Unstacks the mask from the weights in the output tensor for
    segmentation and computes binary accuracy

    Parameters
    ----------
    y_true : Tensor
        Stack of groundtruth segmentation masks + weight maps.
    y_pred : Tensor
        Predicted segmentation masks.

    Returns
    -------
    Tensor
        Binary prediction accuracy.

    '''
    try:
        [seg, weight] = tf.unstack(y_true, 2, axis=-1)

        seg = tf.expand_dims(seg, -1)
        weight = tf.expand_dims(weight, -1)
    except:
        pass    
    
    return keras.metrics.binary_accuracy(seg,y_pred)



#%% Models
# Generic unet declaration:
def unet(input_size = (256,32,1), final_activation = 'sigmoid', output_classes = 1):
    '''
    Generic U-Net declaration.

    Parameters
    ----------
    input_size : tuple of 3 ints, optional
        Dimensions of the input tensor, excluding batch size. 
        The default is (256,32,1).
    final_activation : string or function, optional
        Activation function for the final 2D convolutional layer. see 
        keras.activations
        The default is 'sigmoid'.
    output_classes : int, optional
        Number of output classes, ie dimensionality of the output space of the
        last 2D convolutional layer.
        The default is 1.

    Returns
    -------
    model : Model
        Defined U-Net model (not compiled yet).

    '''
    
    inputs = Input(input_size,  name='true_input')	
	
    conv1 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(inputs)
    conv1 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(conv1)
    conv2 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool1)
    conv2 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(conv2)
    conv3 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool2)
    conv3 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(conv3)
    conv4 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool3)
    conv4 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv4)
    drop4 = Dropout(0.5)(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(drop4)

    conv5 = Conv2D(1024, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool4)
    conv5 = Conv2D(1024, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv5)
    drop5 = Dropout(0.5)(conv5)

    up6 = Conv2D(512, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(drop5))
    merge6 = Concatenate(axis = 3)([drop4,up6]) 
    conv6 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge6)
    conv6 = Conv2D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv6)

    up7 = Conv2D(256, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(conv6))
    merge7 = Concatenate(axis = 3)([conv3,up7])
    conv7 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge7)
    conv7 = Conv2D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv7)

    up8 = Conv2D(128, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(conv7))
    merge8 = Concatenate(axis = 3)([conv2,up8])
    conv8 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge8)
    conv8 = Conv2D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv8)

    up9 = Conv2D(64, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling2D(size = (2,2))(conv8))
    merge9 = Concatenate(axis = 3)([conv1,up9])
    conv9 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge9)
    conv9 = Conv2D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv9)
    conv10 = Conv2D(output_classes, 1, activation = final_activation, name = 'true_output')(conv9)
    
    model = Model(inputs = inputs, outputs = conv10)
    
    return model


# Use the following model for segmentation:
def unet_seg(input_size = (256,32,1)):
    '''
    Cell segmentation U-Net definition function.

    Parameters
    ----------
    input_size : tuple of 3 ints, optional
        Dimensions of the input tensor, without batch size.
        The default is (256,32,1).

    Returns
    -------
    model : Model
        Segmentation U-Net (compiled).

    '''
    
    model = unet(input_size = input_size, final_activation = 'sigmoid', output_classes = 1)
    model.compile(optimizer = Adam(lr = 1e-4), loss = pixelwise_weighted_binary_crossentropy, metrics = [unstack_acc])

    return model


# Use the following model for tracking and lineage reconstruction:
def unet_track(input_size = (256,32,4), class_weights = (1., 1., 1.)):
    '''
    Tracking U-Net definition function.

    Parameters
    ----------
    input_size : tuple of 3 ints, optional
        Dimensions of the input tensor, without batch size.
        The default is (256,32,4).
    class_weights : tuple/list of floats, optional
        Relative weights of each class. If daughters are very uncommon or each
        single cell occupies a small part of the image, this can make the U-Net
        converge to non-trivial solutions. (A trivial solution would be to 
        systematically return a mask of 0s)
        The default is (1., 1., 1.).

    Returns
    -------
    model : Model
        Tracking U-Net (compiled).

    '''
    
    model = unet(input_size = input_size, final_activation = 'softmax', output_classes = 3)
    model.compile(optimizer=Adam(lr = 1e-4), loss=class_weighted_categorical_crossentropy(class_weights), metrics = ['categorical_accuracy'])

    return model


# Use the following model for segmentation:
def unet_chambers(input_size = (512,512,1)):
    '''
    Chambers segmentation U-Net.

    Parameters
    ----------
    input_size : tuple of 3 ints, optional
        Dimensions of the input tensor, without batch size.
        The default is (512,512,1).

    Returns
    -------
    model : Model
        Chambers ID U-Net (compiled).

    '''
    
    model = unet(input_size = input_size, final_activation = 'sigmoid', output_classes = 1)
    model.compile(optimizer = Adam(lr = 1e-4), loss = 'binary_crossentropy')

    return model


'''
This script runs the tracking U-Net on data out of the segmentation U-Net.

Images are processed in batches of 512, although the number of actual samples
run through the tracking U-Net will depend on the number of cells in each 
image.

@author: jblugagne
'''
from data import saveResult_track, predictCompilefromseg_track
from model import unet_track
from os import listdir

# Files:
DeLTA_data = 'C:/DeepLearning/DeLTA_data/'
images_folder = DeLTA_data + 'mother_machine/evaluation/preprocessed/img/'
segmentation_folder = DeLTA_data + 'mother_machine/evaluation/seg_output/'
outputs_folder = DeLTA_data + 'mother_machine/evaluation/track_output/'
model_file = DeLTA_data + 'mother_machine/models/unet_moma_track_multisets.hdf5'
unprocessed = listdir(images_folder)

# Parameters:
target_size = (256, 32)
input_size = target_size + (4,)
process_size = 512 # This is the number of frames to consider _BEFORE_ each sample has been subdivided into cell-specific tracking samples

# Load up model:
model = unet_track(input_size = input_size)
model.load_weights(model_file)

# Process
while(unprocessed):
    # Pop out filenames
    ps = min(process_size,len(unprocessed))
    to_process = unprocessed[0:ps]
    del unprocessed[0:ps]
    
    # Get data:
    inputs, seg_filenames = predictCompilefromseg_track(images_folder, segmentation_folder,
                                                        files_list = to_process, 
                                                        target_size = target_size)
    
    # Predict:
    results = model.predict(inputs,verbose=1)
    
    # Save (use the filenames list from the data compiler)
    saveResult_track(outputs_folder,results, files_list = seg_filenames)


"""
This file can be run in the conda environment described in bioformats_env.yml
It will convert microscopy files like .oif, .nd2, .czi... to an images
sequence on disk that can then be read with the xpreader object in pipeline.py
in the main DeLTA environment (All other files in this repository must be run
in that environment. See delta_env.yml for specs)
Unfortunately bioformats only works in python 2, and our code runs in python 3

It can be run in the command line:
    python bioformats2sequence.py microscopyexperiment.czi /path/to/sequence/folder

Or in an IDE by modifying the paths below

@author: jblugagne
"""

import sys, os
from utilities import xpreader
from cv2 import imwrite

# Files:
if len(sys.argv) == 3: # If arguments were passed
    xpfile = sys.argv[1]
    save_folder = sys.argv[2]
else:
    DeLTA_data = 'C:/DeepLearning/DeLTA_data/'
    xpfile = DeLTA_data + 'mother_machine/evaluation/evaluation_movie.nd2'
    save_folder = DeLTA_data + 'mother_machine/evaluation/sequence/'
if not os.path.exists(save_folder):
    os.mkdir(save_folder)

# Initialize XP reader:
xp = xpreader(xpfile,
              use_bioformats=True)

# Get Filename prototype:
filename_proto = 'Position%0'+str(len(str(xp.positions)))+'d_'\
    +'Channel%0'+str(len(str(xp.channels)))+'d_'\
        +'Frame%0'+str(len(str(xp.timepoints)))+'d'+\
            '.tif'

# Run through frames, save to disk:
counter = 0.
for pos in range(xp.positions):
    for cha in range(xp.channels):
        for fra in range(xp.timepoints):
            # Get frame:
            frame = xp.getframes(pos,cha,fra)
            # Generate filename:
            filename = filename_proto % (pos+1,cha+1,fra+1)
            # Save to disk:
            imwrite(os.path.join(save_folder,filename),frame)
            # display progress:
            counter += 1
            print('%s - saved (%03d%% done)' \
                  % (filename,100*counter/(xp.positions*xp.channels*xp.timepoints)))

# Close the reader:
xp.close()


import numpy as np
from scipy.special import jv

def get_fluorescence_kernel(Lambda,NA,n,radius,scale):
    """
    Returns a 2D numpy array which is an airy-disk approximation of the fluorescence point spread function
    
    Parameters
    ----------
    Lambda : float
        Wavelength of imaging light (micron)
    NA : float
        Numerical aperture of the objective lens
    n : float
        Refractive index of the imaging medium (~1 for air, ~1.4-1.5 for oil)
    radius : int
        The radius of the PSF to be rendered in pixels
    scale : float
        The pixel size of the image to be rendered (micron/pix)
    
    Returns
    -------
    2-D numpy array representing the fluorescence contrast PSF
    """
   

    r = np.arange(-radius,radius+1)
    kaw = 2* NA/n * np.pi/Lambda
    xx,yy = np.meshgrid(r,r)
    xx, yy = xx*scale, yy*scale
    rr = np.sqrt(xx**2+yy**2) * kaw 
    PSF = (2*jv(1,rr)/(rr))**2
    PSF[radius,radius] = 1
    return PSF, np.sqrt(xx**2+yy**2)

        
def somb(x):
    """
    Returns the sombrero function of a 2D numpy array.
    """
    z = np.zeros(x.shape)
    x = np.abs(x)
    idx = np.nonzero(x)
    z[idx] = 2*jv(1,np.pi*x[idx])/(np.pi*x[idx])
    return z



def get_phase_contrast_kernel(R,W,radius,scale,NA,n,sigma,λ):
    """
    Returns a 2D numpy array which is the phase contrast kernel based on microscope parameters
    
    
    Parameters
    ----------
    R : float
        The radius of the phase contrast condenser (in mm)
    W : float
        The width of the phase contrast condenser opening (in mm)
    radius : int
        The radius of the PSF to be rendered in pixels
    scale : float
        The pixel size of the image to be rendered (micron/pix)
    NA : float
        Numerical aperture of the objective lens
    n : float
        Refractive index of the imaging medium (~1 for air, ~1.4-1.5 for oil)
    sigma : radius of a 2D gaussian of the same size as the PSF (in pixels) which is multiplied by the PSF to simulate apodisation of the PSF
    λ : The mean wavelength of the imaging light (in micron)

    Returns
    -------
    2-D numpy array representing the phase contrast PSF
    
    """
    scale1 = 1000 # micron per millimeter
    #F = F * scale1 # to microm
    Lambda = λ # in micron % wavelength of light
    R = R * scale1 # to microm
    W = W * scale1 # to microm
    #The corresponding point spread kernel function for the negative phase contrast 
    r = np.arange(-radius,radius+1)
    xx,yy = np.meshgrid(r,r)
    xx, yy = xx*scale, yy*scale
    kaw = 2* NA/n * np.pi/Lambda
    rr = np.sqrt(xx**2 + yy**2)*kaw

    kernel1 = 2*jv(1,rr)/(rr)
    kernel1[radius,radius] =  1

    kernel2 = 2*(R-W)**2/R**2 * jv(1,(R-W)**2/R**2 * rr)/rr
    kernel2[radius,radius] = np.nanmax(kernel2)

    kernel = kernel1 - kernel2
    kernel = kernel/np.max(kernel)
    kernel[radius,radius] =  1
    kernel = -kernel/np.sum(kernel)
    gaussian = gaussian_2D(radius*2+1, sigma)
    kernel = kernel * gaussian
    return kernel
    
def gaussian_2D(size, σ):
    """Returns a 2D gaussian (numpy array) of size (pixels x pixels) and gaussian radius (σ)"""
    x = np.linspace(0,size,size)
    μ = np.mean(x)
    A = 1/(σ*np.sqrt(2*np.pi))
    B = np.exp(-1/2 * (x-μ)**2/(σ**2))
    _gaussian_1D = A*B
    _gaussian_2D = np.outer(_gaussian_1D,_gaussian_1D)
    return _gaussian_2D

def get_condensers():
    """Returns a dictionary of common phase contrast condenser dimensions, where the numbers are W, R, diameter (in mm)"""
    condensers = {
    "Ph1": (0.45, 3.75, 24),
    "Ph2": (0.8, 5.0, 24),
    "Ph3": (1.0, 9.5, 24),
    "Ph4": (1.5, 14.0, 24),
    "PhF": (1.5, 19.0, 25)
    } #W, R, Diameter
    return condensers